# copyright 2003-2011 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact https://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of CubicWeb.
#
# CubicWeb is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# CubicWeb is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with CubicWeb.  If not, see <https://www.gnu.org/licenses/>.
"""workflow related schemas"""

from yams.buildobjs import (
    EntityType,
    RelationType,
    RelationDefinition,
    SubjectRelation,
    RichString,
    String,
    Int,
)

from cubicweb import _
from cubicweb.schema import RQLConstraint
from cubicweb.schemas import PUB_SYSTEM_ENTITY_PERMS, PUB_SYSTEM_REL_PERMS, RO_REL_PERMS


class Workflow(EntityType):
    __permissions__ = PUB_SYSTEM_ENTITY_PERMS

    name = String(required=True, indexed=True, internationalizable=True, maxsize=256)
    description = RichString(
        default_format="text/rest",
        description=_("semantic description of this workflow"),
    )

    workflow_of = SubjectRelation(
        "CWEType",
        cardinality="+*",
        description=_("entity types which may use this workflow"),
        constraints=[RQLConstraint("O final FALSE")],
    )

    initial_state = SubjectRelation(
        "State",
        cardinality="?*",
        constraints=[
            RQLConstraint(
                "O state_of S", msg=_("state doesn't belong to this workflow")
            )
        ],
        description=_("initial state for this workflow"),
    )


class default_workflow(RelationType):
    """default workflow for an entity type"""

    __permissions__ = PUB_SYSTEM_REL_PERMS

    subject = "CWEType"
    object = "Workflow"
    cardinality = "?*"
    constraints = [
        RQLConstraint(
            "S final FALSE, O workflow_of S",
            msg=_("workflow isn't a workflow for this type"),
        )
    ]


class State(EntityType):
    """used to associate simple states to an entity type and/or to define
    workflows
    """

    __permissions__ = PUB_SYSTEM_ENTITY_PERMS
    __unique_together__ = [("name", "state_of")]
    name = String(required=True, indexed=True, internationalizable=True, maxsize=256)
    description = RichString(
        default_format="text/rest", description=_("semantic description of this state")
    )

    # XXX should be on BaseTransition w/ AND/OR selectors when we will
    # implements #345274
    allowed_transition = SubjectRelation(
        "BaseTransition",
        cardinality="**",
        constraints=[
            RQLConstraint(
                "S state_of WF, O transition_of WF",
                msg=_("state and transition don't belong the the same workflow"),
            )
        ],
        description=_("allowed transitions from this state"),
    )
    state_of = SubjectRelation(
        "Workflow",
        cardinality="1*",
        composite="object",
        inlined=True,
        description=_("workflow to which this state belongs"),
    )


class BaseTransition(EntityType):
    """abstract base class for transitions"""

    __permissions__ = PUB_SYSTEM_ENTITY_PERMS
    __unique_together__ = [("name", "transition_of")]

    name = String(required=True, indexed=True, internationalizable=True, maxsize=256)
    type = String(vocabulary=(_("normal"), _("auto")), default="normal")
    description = RichString(description=_("semantic description of this transition"))

    transition_of = SubjectRelation(
        "Workflow",
        cardinality="1*",
        composite="object",
        inlined=True,
        description=_("workflow to which this transition belongs"),
    )


class require_group(RelationDefinition):
    """group in which a user should be to be allowed to pass this transition"""

    __permissions__ = PUB_SYSTEM_REL_PERMS
    subject = "BaseTransition"
    object = "CWGroup"


class condition(RelationDefinition):
    """a RQL expression which should return some results, else the transition
    won't be available.

    This query may use X and U variables that will respectivly represents the
    current entity and the current user.
    """

    __permissions__ = PUB_SYSTEM_REL_PERMS
    subject = "BaseTransition"
    object = "RQLExpression"
    cardinality = "*?"
    composite = "subject"


class Transition(BaseTransition):
    """use to define a transition from one or multiple states to a destination
    states in workflow's definitions. Transition without destination state will
    go back to the state from which we arrived to the current state.
    """

    __specializes_schema__ = True

    destination_state = SubjectRelation(
        "State",
        cardinality="?*",
        constraints=[
            RQLConstraint(
                "S transition_of WF, O state_of WF",
                msg=_("state and transition don't belong the the same workflow"),
            )
        ],
        description=_("destination state for this transition"),
    )


class WorkflowTransition(BaseTransition):
    """special transition allowing to go through a sub-workflow"""

    __specializes_schema__ = True

    subworkflow = SubjectRelation(
        "Workflow",
        cardinality="1*",
        constraints=[
            RQLConstraint(
                "S transition_of WF, WF workflow_of ET, O workflow_of ET",
                msg=_(
                    "subworkflow isn't a workflow for the same types as the transition's workflow"
                ),
            )
        ],
    )
    # XXX use exit_of and inline it
    subworkflow_exit = SubjectRelation(
        "SubWorkflowExitPoint", cardinality="*1", composite="subject"
    )


class SubWorkflowExitPoint(EntityType):
    """define how we get out from a sub-workflow"""

    subworkflow_state = SubjectRelation(
        "State",
        cardinality="1*",
        constraints=[
            RQLConstraint(
                "T subworkflow_exit S, T subworkflow WF, O state_of WF",
                msg=_("exit state must be a subworkflow state"),
            )
        ],
        description=_("subworkflow state"),
    )
    destination_state = SubjectRelation(
        "State",
        cardinality="?*",
        constraints=[
            RQLConstraint(
                "T subworkflow_exit S, T transition_of WF, O state_of WF",
                msg=_(
                    "destination state must be in the same workflow as our parent transition"
                ),
            )
        ],
        description=_(
            "destination state. No destination state means that transition "
            "should go back to the state from which we've entered the "
            "subworkflow."
        ),
    )


class TrInfo(EntityType):
    """workflow history item"""

    # 'add' security actually done by hooks
    __permissions__ = {
        "read": (
            "managers",
            "users",
            "guests",
        ),  # XXX U has_read_permission O ?
        "add": (
            "managers",
            "users",
            "guests",
        ),
        "delete": (),  # XXX should we allow managers to delete TrInfo?
        "update": (
            "managers",
            "owners",
        ),
    }
    # The unique_together constraint ensures that 2 repositories
    # sharing the db won't be able to fire a transition simultaneously
    # on the same entity tr_count is filled in the FireTransitionHook
    # to the number of TrInfo attached to the entity on which we
    # attempt to fire a transition. In other word, it contains the
    # rank of the TrInfo for that entity, and the constraint says we
    # cannot have 2 TrInfo with the same rank.
    __unique_together__ = [("tr_count", "wf_info_for")]
    from_state = SubjectRelation("State", cardinality="1*", inlined=True)
    to_state = SubjectRelation("State", cardinality="1*", inlined=True)
    # make by_transition optional because we want to allow managers to set
    # entity into an arbitrary state without having to respect wf transition
    by_transition = SubjectRelation("BaseTransition", cardinality="?*")
    comment = RichString(fulltextindexed=True, default_format="text/plain")
    tr_count = Int(
        description="autocomputed attribute used to ensure transition coherency"
    )
    # get actor and date time using owned_by and creation_date


class from_state(RelationType):
    __permissions__ = RO_REL_PERMS.copy()
    inlined = True


class to_state(RelationType):
    __permissions__ = RO_REL_PERMS.copy()
    inlined = True


class by_transition(RelationType):
    # 'add' security actually done by hooks
    __permissions__ = {
        "read": (
            "managers",
            "users",
            "guests",
        ),
        "add": (
            "managers",
            "users",
            "guests",
        ),
        "delete": (),
    }
    inlined = True


class workflow_of(RelationType):
    """link a workflow to one or more entity type"""

    __permissions__ = PUB_SYSTEM_REL_PERMS


class state_of(RelationType):
    """link a state to one or more workflow"""

    __permissions__ = PUB_SYSTEM_REL_PERMS
    inlined = True


class transition_of(RelationType):
    """link a transition to one or more workflow"""

    __permissions__ = PUB_SYSTEM_REL_PERMS
    inlined = True


class destination_state(RelationType):
    """destination state of a transition"""

    __permissions__ = PUB_SYSTEM_REL_PERMS
    inlined = True


class allowed_transition(RelationType):
    """allowed transitions from this state"""

    __permissions__ = PUB_SYSTEM_REL_PERMS


class initial_state(RelationType):
    """indicate which state should be used by default when an entity using
    states is created
    """

    __permissions__ = PUB_SYSTEM_REL_PERMS
    inlined = True


class subworkflow(RelationType):
    __permissions__ = PUB_SYSTEM_REL_PERMS
    inlined = True


class exit_point(RelationType):
    __permissions__ = PUB_SYSTEM_REL_PERMS


class subworkflow_state(RelationType):
    __permissions__ = PUB_SYSTEM_REL_PERMS
    inlined = True


# "abstract" relations, set by WorkflowableEntityType ##########################


class custom_workflow(RelationType):
    """allow to set a specific workflow for an entity"""

    __permissions__ = PUB_SYSTEM_REL_PERMS

    cardinality = "?*"
    constraints = [
        RQLConstraint(
            "S is ET, O workflow_of ET",
            msg=_("workflow isn't a workflow for this type"),
        )
    ]
    object = "Workflow"


class wf_info_for(RelationType):
    """link a transition information to its object"""

    # 'add' security actually done by hooks
    __permissions__ = {
        "read": (
            "managers",
            "users",
            "guests",
        ),
        "add": (
            "managers",
            "users",
            "guests",
        ),
        "delete": (),
    }
    inlined = True

    cardinality = "1*"
    composite = "object"
    fulltext_container = composite
    subject = "TrInfo"


class in_state(RelationType):
    """indicate the current state of an entity"""

    __permissions__ = RO_REL_PERMS

    # not inlined intentionnally since when using ldap sources, user'state
    # has to be stored outside the CWUser table
    inlined = False

    cardinality = "1*"
    constraints = [
        RQLConstraint(
            "S is ET, O state_of WF, WF workflow_of ET",
            msg=_("state doesn't apply to this entity's type"),
        )
    ]
    object = "State"



# copyright 2003-2012 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact https://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of CubicWeb.
#
# CubicWeb is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# CubicWeb is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with CubicWeb.  If not, see <https://www.gnu.org/licenses/>.
"""bookmark related hooks"""


from cubicweb.server import hook


class AutoDeleteBookmarkOp(hook.Operation):
    bookmark = None  # make pylint happy

    def precommit_event(self):
        if not self.cnx.deleted_in_transaction(self.bookmark.eid):
            if not self.bookmark.bookmarked_by:
                self.bookmark.cw_delete()


class DelBookmarkedByHook(hook.Hook):
    """ensure user logins are stripped"""

    __regid__ = "autodelbookmark"
    __select__ = hook.Hook.__select__ & hook.match_rtype(
        "bookmarked_by",
    )
    category = "bookmark"
    events = ("after_delete_relation",)

    def __call__(self):
        AutoDeleteBookmarkOp(self._cw, bookmark=self._cw.entity_from_eid(self.eidfrom))



from yams.buildobjs import EntityType, String, Bytes, RichString


class File(EntityType):
    """a downloadable file which may contains binary data"""

    title = String(fulltextindexed=True, maxsize=256)
    data = Bytes(required=True, description="file to upload")
    data_format = String(
        required=True,
        maxsize=128,
        description=(
            "MIME type of the file. Should be dynamically set " "at upload time."
        ),
    )
    data_encoding = String(
        maxsize=32,
        description=(
            "encoding of the file when it applies (e.g. text). "
            "Should be dynamically set at upload time."
        ),
    )
    data_name = String(
        required=True,
        fulltextindexed=True,
        description=("name of the file. Should be dynamically set " "at upload time."),
    )
    data_hash = String(
        maxsize=256,  # max len of currently available hash alg + prefix is 140
        description=("hash of the file. May be set at upload time."),
        __permissions__={
            "read": ("managers", "users", "guests"),
            "add": (),
            "update": (),
        },
    )
    description = RichString(
        fulltextindexed=True, internationalizable=True, default_format="text/rest"
    )



# copyright 2003-2010 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact https://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of CubicWeb.
#
# CubicWeb is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# CubicWeb is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with CubicWeb.  If not, see <https://www.gnu.org/licenses/>.
"""unit tests for database value generator"""

import datetime
import os.path as osp
import re

from logilab.common.testlib import TestCase, unittest_main

from cubicweb.devtools.apptest_config import ApptestConfiguration
from cubicweb.devtools.fill import ValueGenerator, make_tel

DATADIR = osp.join(osp.abspath(osp.dirname(__file__)), "data")
ISODATE_SRE = re.compile(r"(?P<year>\d{4})-(?P<month>\d{2})-(?P<day>\d{2})$")


class MyValueGenerator(ValueGenerator):
    def generate_Bug_severity(self, entity, index):
        return "dangerous"

    def generate_Any_description(self, entity, index, format=None):
        return "yo"


class ValueGeneratorTC(TestCase):
    """test case for ValueGenerator"""

    def _choice_func(self, etype, attrname):
        try:
            return getattr(self, f"_available_{etype}_{attrname}")(etype, attrname)
        except AttributeError:
            return None

    def _available_Person_firstname(self, etype, attrname):
        return [
            f.strip()
            for f in open(osp.join(DATADIR, "firstnames.txt"), encoding="latin1")
        ]

    def setUp(self):
        config = ApptestConfiguration("data", __file__)
        config.bootstrap_cubes()
        schema = config.load_schema()
        e_schema = schema.entity_schema_for("Person")
        self.person_valgen = ValueGenerator(e_schema, self._choice_func)
        e_schema = schema.entity_schema_for("Bug")
        self.bug_valgen = MyValueGenerator(e_schema)
        self.config = config

    def test_string(self):
        """test string generation"""
        surname = self.person_valgen.generate_attribute_value({}, "surname", 12)
        self.assertEqual(surname, "é&surname12")

    def test_domain_value(self):
        """test value generation from a given domain value"""
        firstname = self.person_valgen.generate_attribute_value({}, "firstname", 12)
        possible_choices = self._choice_func("Person", "firstname")
        self.assertIn(
            firstname, possible_choices, f"{firstname} not in {possible_choices}"
        )

    def test_choice(self):
        """test choice generation"""
        # Test for random index
        for index in range(5):
            sx_value = self.person_valgen.generate_attribute_value(
                {}, "civility", index
            )
            self.assertIn(sx_value, ("Mr", "Mrs", "Ms"))

    def test_integer(self):
        """test integer generation"""
        # Test for random index
        for index in range(5):
            cost_value = self.bug_valgen.generate_attribute_value({}, "cost", index)
            self.assertIn(cost_value, list(range(index + 1)))

    def test_date(self):
        """test date generation"""
        # Test for random index
        for index in range(10):
            date_value = self.person_valgen.generate_attribute_value(
                {}, "birthday", index
            )
            self.assertIsInstance(date_value, datetime.date)

    def test_phone(self):
        """tests make_tel utility"""
        self.assertEqual(make_tel(22030405), "22 03 04 05")

    def test_customized_generation(self):
        self.assertEqual(
            self.bug_valgen.generate_attribute_value({}, "severity", 12), "dangerous"
        )
        self.assertEqual(
            self.bug_valgen.generate_attribute_value({}, "description", 12), "yo"
        )
        self.assertEqual(
            self.person_valgen.generate_attribute_value({}, "description", 12), "yo"
        )


if __name__ == "__main__":
    unittest_main()



# copyright 2003-2014 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact https://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of CubicWeb.
#
# CubicWeb is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# CubicWeb is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with CubicWeb.  If not, see <https://www.gnu.org/licenses/>.
"""server.serverconfig definition"""

import sys
from io import StringIO
from os.path import join

import logilab.common.configuration as lgconfig

from cubicweb.server import SOURCE_TYPES

USER_OPTIONS = (
    (
        "login",
        {
            "type": "string",
            "default": "admin",
            "help": "cubicweb manager account's login " "(this user will be created)",
            "level": 0,
        },
    ),
    (
        "password",
        {
            "type": "password",
            "default": lgconfig.REQUIRED,
            "help": "cubicweb manager account's password",
            "level": 0,
        },
    ),
)


class SourceConfiguration(lgconfig.Configuration):
    def __init__(self, appconfig, options):
        self.appconfig = appconfig  # has to be done before super call
        super().__init__(options=options)

    # make Method('default_instance_id') usable in db option defs (in native.py)
    def default_instance_id(self):
        return self.appconfig.appid

    def input_option(self, option, optdict, inputlevel):
        try:
            dbdriver = self["db-driver"]
        except lgconfig.OptionError:
            pass
        else:
            if dbdriver == "sqlite":
                if option in ("db-user", "db-password"):
                    return
                if option == "db-name":
                    optdict = optdict.copy()
                    optdict["help"] = "path to the sqlite database"
                    optdict["default"] = join(
                        self.appconfig.appdatahome, self.appconfig.appid + ".sqlite"
                    )
        super().input_option(option, optdict, inputlevel)


def ask_source_config(appconfig, type, inputlevel=0):
    options = SOURCE_TYPES[type].options
    sconfig = SourceConfiguration(appconfig, options=options)
    sconfig.input_config(inputlevel=inputlevel)
    return sconfig


def generate_source_config(sconfig, encoding=sys.stdin.encoding):
    """serialize a repository source configuration as text"""
    stream = StringIO()
    optsbysect = list(sconfig.options_by_section())
    assert (
        len(optsbysect) == 1
    ), "all options for a source should be in the same group, got %s" % [
        x[0] for x in optsbysect
    ]
    lgconfig.ini_format(stream, optsbysect[0][1], encoding)
    return stream.getvalue()



# copyright 2017 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# copyright 2014-2016 UNLISH S.A.S. (Montpellier, FRANCE), all rights reserved.
#
# contact https://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of CubicWeb.
#
# CubicWeb is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# CubicWeb is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with CubicWeb.  If not, see <https://www.gnu.org/licenses/>.

"""Binding of CubicWeb connection to Pyramid request."""

import logging

from pyramid import httpexceptions

from cubicweb.pyramid import tools
from cubicweb.server import session as cwsession

log = logging.getLogger(__name__)


class Connection(cwsession.Connection):
    """A specialised Connection that access the session data through a
    property.

    This behavior makes sure the actual session data is not loaded until
    actually accessed.
    """

    def __init__(self, session, *args, **kw):
        super().__init__(session._repo, session._user, *args, **kw)
        self.session = session
        self.lang = session._cached_lang

    def _get_session_data(self):
        return self.session.data

    def _set_session_data(self, data):
        pass

    _session_data = property(_get_session_data, _set_session_data)


class Session:
    """A Session that access the session data through a property.

    Along with :class:`Connection`, it avoid any load of the pyramid session
    data until it is actually accessed.
    """

    def __init__(self, pyramid_request, user, repo):
        self._pyramid_request = pyramid_request
        self._user = user
        self._repo = repo

    @property
    def anonymous_session(self):
        # XXX for now, anonymous_user only exists in webconfig (and testconfig).
        # It will only be present inside all-in-one instance.
        # there is plan to move it down to global config.
        if not hasattr(self._repo.config, "anonymous_user"):
            # not a web or test config, no anonymous user
            return False
        return self._user.login == self._repo.config.anonymous_user()[0]

    def get_data(self):
        if not getattr(self, "_protect_data_access", False):
            self._data_accessed = True
            return self._pyramid_request.session

    def set_data(self, data):
        if getattr(self, "_data_accessed", False):
            self._pyramid_request.session.clear()
            self._pyramid_request.session.update(data)

    data = property(get_data, set_data)

    def new_cnx(self):
        self._protect_data_access = True
        try:
            return Connection(self)
        finally:
            self._protect_data_access = False


def _cw_cnx(request):
    """Obtains a cw session from a pyramid request

    The connection will be commited or rolled-back in a request finish
    callback (this is temporary, we should make use of the transaction manager
    in a later version).

    Not meant for direct use, use ``request.cw_cnx`` instead.

    :param request: A pyramid request
    :returns type: :class:`cubicweb.server.session.Connection`
    """
    session = request.cw_session
    if session is None:
        return None

    cnx = session.new_cnx()

    def commit_state(cnx):
        return cnx.commit_state

    def cleanup(request):
        try:
            if request.exception is not None and not isinstance(
                request.exception,
                (httpexceptions.HTTPSuccessful, httpexceptions.HTTPRedirection),
            ):
                cnx.rollback()
            elif commit_state(cnx) == "uncommitable":
                cnx.rollback()
            else:
                cnx.commit()
        finally:
            cnx.__exit__(None, None, None)

    request.add_finished_callback(cleanup)
    cnx.__enter__()
    return cnx


def repo_connect(request, repo, eid):
    """A lightweight version of
    :meth:`cubicweb.server.repository.Repository.connect` that does not keep
    track of opened sessions, removing the need of closing them"""
    user, lang = tools.cached_build_user(repo, eid)
    session = Session(request, user, repo)
    session._cached_lang = lang
    tools.cnx_attach_entity(session, user)
    return session


def _cw_session(request):
    """Obtains a cw session from a pyramid request

    :param request: A pyramid request
    :returns type: :class:`cubicweb.server.session.Session`

    Not meant for direct use, use ``request.cw_session`` instead.
    """
    repo = request.registry["cubicweb.repository"]

    if not request.authenticated_userid:
        eid = request.registry.get("cubicweb.anonymous_eid")
        if eid is None:
            return None
        session = repo_connect(request, repo, eid=eid)
    else:
        session = request._cw_cached_session

    return session


def get_principals(login, request):
    """Returns the group names of the authenticated user.

    This function is meant to be used as an authentication policy callback.

    It also pre-open the cubicweb session and put it in
    request._cw_cached_session for later usage by :func:`_cw_session`.

    .. note::

        If the default authentication policy is not used, make sure this
        function gets called by the active authentication policy.

    :param login: A cubicweb user eid
    :param request: A pyramid request
    :returns: A list of group names
    """
    repo = request.registry["cubicweb.repository"]

    try:
        session = repo_connect(request, repo, eid=login)
        request._cw_cached_session = session
    except Exception:
        log.exception("Failed")
        raise

    with session.new_cnx() as cnx:
        with cnx.security_enabled(read=False):
            return {
                group
                for group, in cnx.execute(
                    "Any GN WHERE U in_group G, G name GN, U eid %(userid)s",
                    {"userid": login},
                )
            }


def includeme(config):
    """Enables the core features of Pyramid CubicWeb.

    Automatically called by the 'pyramid' command, or via
    ``config.include('cubicweb.pyramid.code')``. In the later case,
    the following registry entries must be defined first:

    'cubicweb.config'
        A cubicweb 'config' instance.

    'cubicweb.repository'
        The correponding cubicweb repository.

    'cubicweb.registry'
        The vreg.
    """
    repo = config.registry["cubicweb.repository"]

    with repo.internal_cnx() as cnx:
        login = config.registry["cubicweb.config"].anonymous_user()[0]
        if login is not None:
            config.registry["cubicweb.anonymous_eid"] = (
                cnx.find("CWUser", login=login).one().eid
            )

    config.add_request_method(_cw_session, name="cw_session", property=True, reify=True)
    config.add_request_method(_cw_cnx, name="cw_cnx", property=True, reify=True)

    cwcfg = config.registry["cubicweb.config"]
    for cube in cwcfg.cubes():
        pkgname = f"cubicweb_{cube}"
        mod = __import__(pkgname)
        if hasattr(mod, "includeme"):
            config.include(pkgname)



# copyright 2017 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact https://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of CubicWeb.
#
# CubicWeb is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# CubicWeb is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with CubicWeb.  If not, see <https://www.gnu.org/licenses/>.

import copy

from cubicweb.devtools import testlib
from cubicweb.wfutils import setup_workflow


class TestWFUtils(testlib.CubicWebTC):
    defs = {
        "group": {
            "etypes": "CWGroup",
            "default": True,
            "initial_state": "draft",
            "states": ["draft", "published"],
            "transitions": {
                "publish": {
                    "fromstates": "draft",
                    "tostate": "published",
                    "requiredgroups": "managers",
                }
            },
        }
    }

    def test_create_workflow(self):
        with self.admin_access.cnx() as cnx:
            wf = setup_workflow(cnx, "group", self.defs["group"])
            self.assertEqual(wf.name, "group")
            self.assertEqual(wf.initial.name, "draft")

            draft = wf.state_by_name("draft")
            self.assertIsNotNone(draft)

            published = wf.state_by_name("published")
            self.assertIsNotNone(published)

            publish = wf.transition_by_name("publish")
            self.assertIsNotNone(publish)

            self.assertEqual(publish.destination_state, (published,))
            self.assertEqual(draft.allowed_transition, (publish,))

            self.assertEqual({g.name for g in publish.require_group}, {"managers"})

    def test_update(self):
        with self.admin_access.cnx() as cnx:
            wf = setup_workflow(cnx, "group", self.defs["group"])
            eid = wf.eid

        with self.admin_access.cnx() as cnx:
            wfdef = copy.deepcopy(self.defs["group"])
            wfdef["states"].append("new")
            wfdef["initial_state"] = "new"
            wfdef["transitions"]["publish"]["fromstates"] = ("draft", "new")
            wfdef["transitions"]["publish"]["requiredgroups"] = ("managers", "users")
            wfdef["transitions"]["todraft"] = {
                "fromstates": ("new", "published"),
                "tostate": "draft",
            }

            wf = setup_workflow(cnx, "group", wfdef)
            self.assertEqual(wf.eid, eid)
            self.assertEqual(wf.name, "group")
            self.assertEqual(wf.initial.name, "new")

            new = wf.state_by_name("new")
            self.assertIsNotNone(new)

            draft = wf.state_by_name("draft")
            self.assertIsNotNone(draft)

            published = wf.state_by_name("published")
            self.assertIsNotNone(published)

            publish = wf.transition_by_name("publish")
            self.assertIsNotNone(publish)

            todraft = wf.transition_by_name("todraft")
            self.assertIsNotNone(todraft)

            self.assertEqual(
                {g.name for g in publish.require_group}, {"managers", "users"}
            )

            self.assertEqual(publish.destination_state, (published,))
            self.assertEqual(draft.allowed_transition, (publish,))
            self.assertEqual(todraft.destination_state, (draft,))
            self.assertEqual(published.allowed_transition, (todraft,))
            self.assertCountEqual(new.allowed_transition, (publish, todraft))


if __name__ == "__main__":
    import unittest

    unittest.main()



# copyright 2003-2016 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact https://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of CubicWeb.
#
# CubicWeb is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# CubicWeb is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with CubicWeb.  If not, see <https://www.gnu.org/licenses/>.
"""Core hooks: set generic metadata"""

from datetime import datetime

from pytz import utc

from cubicweb.predicates import is_instance
from cubicweb.server import hook


class MetaDataHook(hook.Hook):
    __abstract__ = True
    category = "metadata"


class InitMetaAttrsHook(MetaDataHook):
    """before create a new entity -> set creation and modification date

    this is a conveniency hook, you shouldn't have to disable it
    """

    __regid__ = "metaattrsinit"
    events = ("before_add_entity",)

    def __call__(self):
        timestamp = datetime.now(utc)
        edited = self.entity.cw_edited
        if not edited.get("creation_date"):
            edited["creation_date"] = timestamp
        if not edited.get("modification_date"):
            edited["modification_date"] = timestamp
        if not self._cw.transaction_data.get("do-not-insert-cwuri"):
            cwuri = f"{self._cw.base_url()}{self.entity.eid}"
            edited.setdefault("cwuri", cwuri)


class UpdateMetaAttrsHook(MetaDataHook):
    """update an entity -> set modification date"""

    __regid__ = "metaattrsupdate"
    events = ("before_update_entity",)

    def __call__(self):
        # repairing is true during c-c upgrade/shell and similar commands. We
        # usually don't want to update modification date in such cases.
        #
        # XXX to be really clean, we should turn off modification_date update
        # explicitly on each command where we do not want that behaviour.
        if not self._cw.vreg.config.repairing:
            self.entity.cw_edited.setdefault("modification_date", datetime.now(utc))


class SetCreatorOp(hook.DataOperationMixIn, hook.Operation):
    def precommit_event(self):
        cnx = self.cnx
        relations = [
            (eid, cnx.user.eid)
            for eid in self.get_data()
            # don't consider entities that have been created and deleted in
            # the same transaction, nor ones where created_by has been
            # explicitly set
            if not cnx.deleted_in_transaction(eid)
            and not cnx.entity_from_eid(eid).created_by
        ]
        cnx.add_relations([("created_by", relations)])


class SetOwnershipHook(MetaDataHook):
    """create a new entity -> set owner and creator metadata"""

    __regid__ = "setowner"
    events = ("after_add_entity",)

    def __call__(self):
        if not self._cw.is_internal_session:
            self._cw.add_relation(self.entity.eid, "owned_by", self._cw.user.eid)
            SetCreatorOp.get_instance(self._cw).add_data(self.entity.eid)


class SyncOwnersOp(hook.DataOperationMixIn, hook.Operation):
    def precommit_event(self):
        for compositeeid, composedeid in self.get_data():
            if self.cnx.deleted_in_transaction(compositeeid):
                continue
            if self.cnx.deleted_in_transaction(composedeid):
                continue
            self.cnx.execute(
                "SET X owned_by U WHERE C owned_by U, C eid %(c)s,"
                "NOT EXISTS(X owned_by U, X eid %(x)s)",
                {"c": compositeeid, "x": composedeid},
            )


class SyncCompositeOwner(MetaDataHook):
    """when adding composite relation, the composed should have the same owners
    has the composite
    """

    __regid__ = "synccompositeowner"
    events = ("after_add_relation",)

    def __call__(self):
        if self.rtype == "wf_info_for":
            # skip this special composite relation # XXX (syt) why?
            return
        eidfrom, eidto = self.eidfrom, self.eidto
        composite = self._cw.rtype_eids_rdef(self.rtype, eidfrom, eidto).composite
        if composite == "subject":
            SyncOwnersOp.get_instance(self._cw).add_data((eidfrom, eidto))
        elif composite == "object":
            SyncOwnersOp.get_instance(self._cw).add_data((eidto, eidfrom))


class FixUserOwnershipHook(MetaDataHook):
    """when a user has been created, add owned_by relation on itself"""

    __regid__ = "fixuserowner"
    __select__ = MetaDataHook.__select__ & is_instance("CWUser")
    events = ("after_add_entity",)

    def __call__(self):
        self._cw.add_relation(self.entity.eid, "owned_by", self.entity.eid)


class UpdateFTIHook(MetaDataHook):
    """sync fulltext index text index container when a relation with
    fulltext_container set is added / removed
    """

    __regid__ = "updateftirel"
    events = ("after_add_relation", "after_delete_relation")

    def __call__(self):
        rtype = self.rtype
        cnx = self._cw
        ftcontainer = cnx.vreg.schema.relation_schema_for(rtype).fulltext_container
        if ftcontainer == "subject":
            cnx.repo.system_source.index_entity(cnx, cnx.entity_from_eid(self.eidfrom))
        elif ftcontainer == "object":
            cnx.repo.system_source.index_entity(cnx, cnx.entity_from_eid(self.eidto))



from yams.buildobjs import EntityType, String, RichString


class Card(EntityType):
    __permissions__ = {
        "read": ("managers", "users", "guests"),
        "add": ("managers", "users"),
        "delete": ("managers", "owners"),
        "update": (
            "managers",
            "owners",
        ),
    }

    title = String(required=True, fulltextindexed=True, maxsize=256)
    synopsis = String(
        fulltextindexed=True, maxsize=512, description=("an abstract for this card")
    )
    content = RichString(
        fulltextindexed=True, internationalizable=True, default_format="text/rest"
    )
    wikiid = String(maxsize=64, unique=True)



# copyright 2003-2015 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact https://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of CubicWeb.
#
# CubicWeb is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# CubicWeb is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with CubicWeb.  If not, see <https://www.gnu.org/licenses/>.
"""some hooks to handle notification on entity's changes"""


from logilab.common.textutils import normalize_text
from logilab.common.registry import RegistryNotFound

from cubicweb.predicates import is_instance
from cubicweb.server import hook
from cubicweb.sobjects.supervising import SupervisionMailOp


def notify_on_commit(cnx, view, viewargs=None):
    """register a notification view (see
    :class:`~cubicweb.sobjects.notification.NotificationView`) to be sent at
    post-commit time, ie only if the transaction has succeeded.

    `viewargs` is an optional dictionary containing extra argument to be given
    to :meth:`~cubicweb.sobjects.notification.NotificationView.render_and_send`
    """
    if viewargs is None:
        viewargs = {}
    notif_op = _RenderAndSendNotificationOp.get_instance(cnx)
    notif_op.add_data((view, viewargs))


class _RenderAndSendNotificationOp(hook.DataOperationMixIn, hook.Operation):
    """End of the notification chain. Do render and send views after commit

    All others Operations end up adding data to this Operation.
    The notification are done on ``postcommit_event`` to make sure to prevent
    sending notification about rolled back data.
    """

    containercls = list

    def postcommit_event(self):
        deleted = self.cnx.deleted_in_transaction
        views_to_render = self.get_data()
        for view, viewargs in views_to_render:
            if view.cw_rset is not None:
                if not view.cw_rset:
                    # entity added and deleted in the same transaction
                    # (cache effect)
                    continue
                elif deleted(view.cw_rset[view.cw_row or 0][view.cw_col or 0]):
                    # entity added and deleted in the same transaction
                    continue
            try:
                view.render_and_send(**viewargs)
            except Exception:
                # error in post commit are not propagated
                # We keep this logic here to prevent a small notification error
                # to prevent them all.
                self.exception("Notification failed")

                if self.cnx.vreg.config.mode == "test":
                    # reraise in testing context because we actually want to
                    # have those exceptions here and that self.exception is
                    # filtered in test context
                    raise


class NotificationHook(hook.Hook):
    __abstract__ = True
    category = "notification"

    def select_view(self, vid, rset, row=0, col=0):
        try:
            return self._cw.vreg["notification-views"].select_or_none(
                vid, self._cw, rset=rset, row=row, col=col
            )
        except RegistryNotFound:  # can happen in some config
            # (e.g. repo only config with no
            # notification views registered by
            # the instance's cubes)
            return None


class StatusChangeHook(NotificationHook):
    """notify when a workflowable entity has its state modified"""

    __regid__ = "notifystatuschange"
    __select__ = NotificationHook.__select__ & is_instance("TrInfo")
    events = ("after_add_entity",)

    def __call__(self):
        entity = self.entity
        if not entity.from_state:  # not a transition
            return
        rset = entity.related("wf_info_for")
        view = self.select_view("notif_status_change", rset=rset, row=0)
        if view is None:
            return
        comment = entity.printable_value("comment", format="text/plain")
        # XXX don't try to wrap rest until we've a proper transformation (see
        # #103822)
        if comment and entity.comment_format != "text/rest":
            comment = normalize_text(comment, 80)
        viewargs = {
            "comment": comment,
            "previous_state": entity.previous_state.name,
            "current_state": entity.new_state.name,
        }
        notify_on_commit(self._cw, view, viewargs=viewargs)


class RelationChangeHook(NotificationHook):
    __regid__ = "notifyrelationchange"
    events = (
        "before_add_relation",
        "after_add_relation",
        "before_delete_relation",
        "after_delete_relation",
    )

    def __call__(self):
        """if a notification view is defined for the event, send notification
        email defined by the view
        """
        rset = self._cw.eid_rset(self.eidfrom)
        view = self.select_view(f"notif_{self.event}_{self.rtype}", rset=rset, row=0)
        if view is None:
            return
        notify_on_commit(self._cw, view)


class EntityChangeHook(NotificationHook):
    """if a notification view is defined for the event, send notification
    email defined by the view
    """

    __regid__ = "notifyentitychange"
    events = ("after_add_entity", "after_update_entity")

    def __call__(self):
        rset = self.entity.as_rset()
        view = self.select_view(f"notif_{self.event}", rset=rset, row=0)
        if view is None:
            return
        notify_on_commit(self._cw, view)


class EntityUpdatedNotificationOp(hook.SingleLastOperation):
    """scrap all changed entity to prepare a Notification Operation for them"""

    def precommit_event(self):
        # precommit event that creates postcommit operation
        cnx = self.cnx
        for eid in cnx.transaction_data["changes"]:
            view = cnx.vreg["notification-views"].select(
                "notif_entity_updated", cnx, rset=cnx.eid_rset(eid), row=0
            )
            notify_on_commit(
                self.cnx,
                view,
                viewargs={"changes": cnx.transaction_data["changes"][eid]},
            )


class EntityUpdateHook(NotificationHook):
    __regid__ = "notifentityupdated"
    __abstract__ = True  # do not register by default
    __select__ = NotificationHook.__select__ & hook.issued_from_user_query()
    events = ("before_update_entity",)
    skip_attrs = {"modification_date"}

    def __call__(self):
        cnx = self._cw
        if cnx.added_in_transaction(self.entity.eid):
            return  # entity is being created
        # then compute changes
        attrs = [k for k in self.entity.cw_edited if k not in self.skip_attrs]
        if not attrs:
            return
        changes = cnx.transaction_data.setdefault("changes", {})
        thisentitychanges = changes.setdefault(self.entity.eid, set())
        rqlsel, rqlrestr = [], ["X eid %(x)s"]
        for i, attr in enumerate(attrs):
            var = chr(65 + i)
            rqlsel.append(var)
            rqlrestr.append(f"X {attr} {var}")
        rql = f"Any {','.join(rqlsel)} WHERE {','.join(rqlrestr)}"
        rset = cnx.execute(rql, {"x": self.entity.eid})
        for i, attr in enumerate(attrs):
            oldvalue = rset[0][i]
            newvalue = self.entity.cw_edited[attr]
            if oldvalue != newvalue:
                thisentitychanges.add((attr, oldvalue, newvalue))
        if thisentitychanges:
            EntityUpdatedNotificationOp(cnx)


# supervising ##################################################################


class SomethingChangedHook(NotificationHook):
    __regid__ = "supervising"
    __select__ = NotificationHook.__select__ & hook.issued_from_user_query()
    events = (
        "before_add_relation",
        "before_delete_relation",
        "after_add_entity",
        "before_update_entity",
    )

    def __call__(self):
        dest = self._cw.vreg.config["supervising-addrs"]
        if not dest:  # no supervisors, don't do this for nothing...
            return
        if self._call():
            SupervisionMailOp(self._cw)

    def _call(self):
        event = self.event.split("_", 1)[1]
        if event == "update_entity":
            if self._cw.added_in_transaction(self.entity.eid):
                return False
            if self.entity.e_schema == "CWUser":
                if not (
                    frozenset(self.entity.cw_edited)
                    - frozenset(("eid", "modification_date", "last_login_time"))
                ):
                    # don't record last_login_time update which are done
                    # automatically at login time
                    return False
        self._cw.transaction_data.setdefault("pendingchanges", []).append((event, self))
        return True


class EntityDeleteHook(SomethingChangedHook):
    __regid__ = "supervisingentitydel"
    events = ("before_delete_entity",)

    def _call(self):
        try:
            title = self.entity.dc_title()
        except Exception:
            # may raise an error during deletion process, for instance due to
            # missing required relation
            title = f"#{self.entity.eid}"
        self._cw.transaction_data.setdefault("pendingchanges", []).append(
            ("delete_entity", (self.entity.eid, self.entity.cw_etype, title))
        )
        return True



# copyright 2003-2016 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact https://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of CubicWeb.
#
# CubicWeb is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# CubicWeb is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with CubicWeb.  If not, see <https://www.gnu.org/licenses/>.
"""functions for schema / permissions (de)serialization using RQL"""

import json
import sqlite3
import sys

from logilab.common.shellutils import ProgressBar, DummyProgressBar
from yams import BadSchemaDefinition, schema as schemamod, buildobjs as ybo, constraints

from cubicweb import Binary, ETYPE_NAME_MAP
from cubicweb.schema import KNOWN_RPROPERTIES, CONSTRAINTS, VIRTUAL_RTYPES
from cubicweb.server import sqlutils, schema2sql as y2sql


def group_mapping(cnx, interactive=True):
    """create a group mapping from an rql cursor

    A group mapping has standard group names as key (managers, owners at least)
    and the actual CWGroup entity's eid as associated value.
    In interactive mode (the default), missing groups'eid will be prompted
    from the user.
    """
    res = {}
    for eid, name in cnx.execute(
        "Any G, N WHERE G is CWGroup, G name N", build_descr=False
    ):
        res[name] = eid
    if not interactive:
        return res
    missing = [g for g in ("owners", "managers", "users", "guests") if g not in res]
    if missing:
        print(
            "some native groups are missing but the following groups have been found:"
        )
        print("\n".join(f"* {n} ({eid})" for n, eid in res.items()))
        print()
        print("enter the eid of a to group to map to each missing native group")
        print("or just type enter to skip permissions granted to a group")
        for group in missing:
            while True:
                value = input(f"eid for group {group}: ").strip()
                if not value:
                    continue
                try:
                    eid = int(value)
                except ValueError:
                    print("eid should be an integer")
                    continue
                for eid_ in res.values():
                    if eid == eid_:
                        break
                else:
                    print("eid is not a group eid")
                    continue
                res[name] = eid
                break
    return res


def cstrtype_mapping(cnx):
    """cached constraint types mapping"""
    map = dict(cnx.execute("Any T, X WHERE X is CWConstraintType, X name T"))
    return map


# schema / perms deserialization ##############################################


def deserialize_schema(schema, cnx):
    """return a schema according to information stored in an rql database
    as CWRType and CWEType entities
    """
    repo = cnx.repo
    dbhelper = repo.system_source.dbhelper

    # Computed Rtype
    tables = set(t.lower() for t in dbhelper.list_tables(cnx.cnxset.cu))
    # on postgres, ensure the default schema name (public) is also checked
    if dbhelper.backend_name == "postgres":
        tables.update(t.lower() for t in dbhelper.list_tables(cnx.cnxset.cu, "public"))
    has_computed_relations = "cw_cwcomputedrtype" in tables
    # computed attribute
    try:
        cnx.system_sql("SELECT cw_formula FROM cw_CWAttribute")
        has_computed_attributes = True
    except Exception:
        cnx.rollback()
        has_computed_attributes = False

    # XXX bw compat (3.6 migration)
    sqlcu = cnx.system_sql("SELECT * FROM cw_CWRType WHERE cw_name='symetric'")
    if sqlcu.fetchall():
        sql = dbhelper.sql_rename_col(
            "cw_CWRType",
            "cw_symetric",
            "cw_symmetric",
            dbhelper.TYPE_MAPPING["Boolean"],
            True,
        )
        sqlcu.execute(sql)
        sqlcu.execute(
            "UPDATE cw_CWRType SET cw_name='symmetric' WHERE cw_name='symetric'"
        )
        cnx.commit()
    ertidx = {}
    copiedeids = set()
    permsidx = deserialize_ertype_permissions(cnx)
    schema.reading_from_database = True
    # load every entity types
    for eid, etype, desc in cnx.execute(
        "Any X, N, D WHERE X is CWEType, X name N, X description D", build_descr=False
    ):
        # base types are already in the schema, skip them
        if etype in schemamod.BASE_TYPES:
            # just set the eid
            eschema = schema.entity_schema_for(etype)
            eschema.eid = eid
            ertidx[eid] = etype
            continue
        if etype in ETYPE_NAME_MAP:
            needcopy = False
            netype = ETYPE_NAME_MAP[etype]
            # can't use write rql queries at this point, use raw sql
            sqlexec = cnx.system_sql
            if sqlexec(
                "SELECT 1 FROM %(p)sCWEType WHERE %(p)sname=%%(n)s"
                % {"p": sqlutils.SQL_PREFIX},
                {"n": netype},
            ).fetchone():
                # the new type already exists, we should copy (eg make existing
                # instances of the old type instances of the new type)
                assert etype.lower() != netype.lower()
                needcopy = True
            else:
                # the new type doesn't exist, we should rename
                sqlexec(
                    "UPDATE %(p)sCWEType SET %(p)sname=%%(n)s WHERE %(p)seid=%%(x)s"
                    % {"p": sqlutils.SQL_PREFIX},
                    {"x": eid, "n": netype},
                )
                if etype.lower() != netype.lower():
                    alter_table_sql = dbhelper.sql_rename_table(
                        sqlutils.SQL_PREFIX + etype, sqlutils.SQL_PREFIX + netype
                    )
                    sqlexec(alter_table_sql)
            sqlexec(
                "UPDATE entities SET type=%(n)s WHERE type=%(x)s",
                {"x": etype, "n": netype},
            )
            cnx.commit(False)
            tocleanup = [eid]
            tocleanup += (
                eid for eid, cached in repo._type_cache.items() if etype == cached
            )
            repo.clear_caches(tocleanup)
            cnx.commit(False)
            if needcopy:
                ertidx[eid] = netype
                copiedeids.add(eid)
                # copy / CWEType entity removal expected to be done through
                # rename_entity_type in a migration script
                continue
            etype = netype
        ertidx[eid] = etype
        eschema = schema.add_entity_type(
            ybo.EntityType(name=etype, description=desc, eid=eid)
        )
        set_perms(eschema, permsidx)
    # load inheritance relations
    for etype, stype in cnx.execute(
        "Any XN, ETN WHERE X is CWEType, X name XN, X specializes ET, ET name ETN",
        build_descr=False,
    ):
        etype = ETYPE_NAME_MAP.get(etype, etype)
        stype = ETYPE_NAME_MAP.get(stype, stype)
        schema.entity_schema_for(etype)._specialized_type = stype
        schema.entity_schema_for(stype)._specialized_by.append(etype)
    if has_computed_relations:
        rset = cnx.execute(
            "Any X, N, R, D WHERE X is CWComputedRType, X name N, "
            "X rule R, X description D"
        )
        for eid, rule_name, rule, description in rset.rows:
            rtype = ybo.ComputedRelation(
                name=rule_name, rule=rule, eid=eid, description=description
            )
            rschema = schema.add_relation_type(rtype)
            set_perms(rschema, permsidx)
    # load every relation types
    for eid, rtype, desc, sym, il, ftc in cnx.execute(
        "Any X,N,D,S,I,FTC WHERE X is CWRType, X name N, X description D, "
        "X symmetric S, X inlined I, X fulltext_container FTC",
        build_descr=False,
    ):
        ertidx[eid] = rtype
        rschema = schema.add_relation_type(
            ybo.RelationType(
                name=rtype,
                description=desc,
                symmetric=bool(sym),
                inlined=bool(il),
                fulltext_container=ftc,
                eid=eid,
            )
        )
    # remains to load every relation definitions (ie relations and attributes)
    cstrsidx = deserialize_rdef_constraints(cnx)
    pendingrdefs = []

    # closure to factorize common code of attribute/relation rdef addition

    def _add_rdef(rdefeid, seid, reid, oeid, **kwargs):
        rdef = ybo.RelationDefinition(
            ertidx[seid],
            ertidx[reid],
            ertidx[oeid],
            constraints=cstrsidx.get(rdefeid, ()),
            eid=rdefeid,
            **kwargs,
        )
        if seid in copiedeids or oeid in copiedeids:
            # delay addition of this rdef. We'll insert them later if needed. We
            # have to do this because:
            #
            # * on etype renaming, we want relation of the old entity type being
            #   redirected to the new type during migration
            #
            # * in the case of a copy, we've to take care that rdef already
            #   existing in the schema are not overwritten by a redirected one,
            #   since we want correct eid on them (redirected rdef will be
            #   removed in rename_entity_type)
            pendingrdefs.append(rdef)
        else:
            # add_relation_def return a RelationDefinitionSchema if it has been
            # actually added (can be None on duplicated relation definitions,
            # e.g. if the relation type is marked as beeing symmetric)
            rdefs = schema.add_relation_def(rdef)
            if rdefs is not None:
                ertidx[rdefeid] = rdefs
                set_perms(rdefs, permsidx)

    # Get the type parameters for additional base types.
    try:
        extra_props = dict(
            cnx.execute("Any X, XTP WHERE X is CWAttribute, " "X extra_props XTP")
        )
    except Exception:
        cnx.critical(
            "Previous CRITICAL notification about extra_props is not "
            "a problem if you are migrating to cubicweb 3.17"
        )
        extra_props = {}  # not yet in the schema (introduced by 3.17 migration)

    # load attributes
    rql = (
        "Any X,SE,RT,OE,CARD,ORD,DESC,IDX,FTIDX,I18N,DFLT%(fm)s "
        "WHERE X is CWAttribute, X relation_type RT, X cardinality CARD,"
        "      X ordernum ORD, X indexed IDX, X description DESC, "
        "      X internationalizable I18N, X defaultval DFLT,%(fmsnip)s"
        "      X fulltextindexed FTIDX, X from_entity SE, X to_entity OE"
    )
    if has_computed_attributes:
        rql = rql % {"fm": ",FM", "fmsnip": "X formula FM,"}
    else:
        rql = rql % {"fm": "", "fmsnip": ""}
    for values in cnx.execute(rql, build_descr=False):
        attrs = dict(
            zip(
                (
                    "rdefeid",
                    "seid",
                    "reid",
                    "oeid",
                    "cardinality",
                    "order",
                    "description",
                    "indexed",
                    "fulltextindexed",
                    "internationalizable",
                    "default",
                    "formula",
                ),
                values,
            )
        )
        typeparams = extra_props.get(attrs["rdefeid"])
        attrs.update(
            json.loads(typeparams.getvalue().decode("ascii")) if typeparams else {}
        )
        default = attrs["default"]
        if default is not None:
            if isinstance(default, Binary):
                # while migrating from 3.17 to 3.18, we still have to
                # handle String defaults
                attrs["default"] = default.unzpickle()
        _add_rdef(**attrs)
    # load relations
    for values in cnx.execute(
        "Any X,SE,RT,OE,CARD,ORD,DESC,C WHERE X is CWRelation, X relation_type RT,"
        "X cardinality CARD, X ordernum ORD, X description DESC, "
        "X from_entity SE, X to_entity OE, X composite C",
        build_descr=False,
    ):
        rdefeid, seid, reid, oeid, card, ord, desc, comp = values
        _add_rdef(
            rdefeid,
            seid,
            reid,
            oeid,
            cardinality=card,
            description=desc,
            order=ord,
            composite=comp,
        )
    for rdef in pendingrdefs:
        try:
            rdefs = schema.add_relation_def(rdef)
        except BadSchemaDefinition:
            continue
        if rdefs is not None:
            set_perms(rdefs, permsidx)
    unique_togethers = {}
    rset = cnx.execute(
        "Any X,E,R WHERE "
        "X is CWUniqueTogetherConstraint, "
        "X constraint_of E, X relations R",
        build_descr=False,
    )
    for values in rset:
        uniquecstreid, eeid, releid = values
        eschema = schema.schema_by_eid(eeid)
        relations = unique_togethers.setdefault(uniquecstreid, (eschema, []))
        rel = ertidx[releid]
        if isinstance(rel, schemamod.RelationDefinitionSchema):
            # not yet migrated 3.9 database ('relations' target type changed
            # to CWRType in 3.10)
            rtype = rel.rtype.type
        else:
            rtype = str(rel)
        relations[1].append(rtype)
    for eschema, unique_together in unique_togethers.values():
        eschema._unique_together.append(tuple(sorted(unique_together)))
    schema.infer_specialization_rules()
    cnx.commit()
    schema.finalize()
    schema.reading_from_database = False


def deserialize_ertype_permissions(cnx):
    """return sect action:groups associations for the given
    entity or relation schema with its eid, according to schema's
    permissions stored in the database as [read|add|delete|update]_permission
    relations between CWEType/CWRType and CWGroup entities
    """
    res = {}
    for action in ("read", "add", "update", "delete"):
        rql = f"Any E,N WHERE G is CWGroup, G name N, E {action}_permission G"
        for eid, gname in cnx.execute(rql, build_descr=False):
            res.setdefault(eid, {}).setdefault(action, []).append(gname)
        rql = (
            "Any E,X,EXPR,V WHERE X is RQLExpression, X expression EXPR, "
            "E %s_permission X, X mainvars V" % action
        )
        for eid, expreid, expr, mainvars in cnx.execute(rql, build_descr=False):
            # we don't know yet if it's a rql expr for an entity or a relation,
            # so append a tuple to differentiate from groups and so we'll be
            # able to instantiate it later
            res.setdefault(eid, {}).setdefault(action, []).append(
                (expr, mainvars, expreid)
            )
    return res


def deserialize_rdef_constraints(cnx):
    """return the list of relation definition's constraints as instances"""
    if cnx.repo.system_source.dbdriver != "sqlite" or sqlite3.sqlite_version_info >= (
        3,
        7,
        12,
    ):
        # these are implemented as CHECK constraints in sql, don't do the work twice. Unless we
        # are using too old version of sqlite which misses the constraint name in the integrity
        # error so we've to check them by ourselves anyway
        constraints.StaticVocabularyConstraint.check = lambda *args: True
        constraints.IntervalBoundConstraint.check = lambda *args: True
        constraints.BoundaryConstraint.check = lambda *args: True
    res = {}
    for rdefeid, ceid, ct, val in cnx.execute(
        "Any E, X,TN,V WHERE E constrained_by X, X is CWConstraint, "
        "X cstrtype T, T name TN, X value V",
        build_descr=False,
    ):
        cstr = CONSTRAINTS[ct].deserialize(val)
        cstr.eid = ceid
        res.setdefault(rdefeid, []).append(cstr)
    return res


def set_perms(erschema, permsidx):
    """set permissions on the given erschema according to the permission
    definition dictionary as built by deserialize_ertype_permissions for a
    given erschema's eid
    """
    # reset erschema permissions here to avoid getting yams default anyway
    erschema.permissions = dict((action, ()) for action in erschema.ACTIONS)
    try:
        thispermsdict = permsidx[erschema.eid]
    except KeyError:
        return
    for action, somethings in thispermsdict.items():
        erschema.permissions[action] = tuple(
            isinstance(p, tuple) and erschema.rql_expression(*p) or p
            for p in somethings
        )


# schema / perms serialization ################################################


def serialize_schema(cnx, schema):
    """synchronize schema and permissions in the database according to
    current schema
    """
    _title = "-> storing the schema in the database "
    print(_title, end=" ")
    execute = cnx.execute
    eschemas = schema.entities()
    pb_size = (
        len(eschemas + schema.relations())
        + len(CONSTRAINTS)
        + len([x for x in eschemas if x.specializes()])
    )
    if sys.stdout.isatty():
        pb = ProgressBar(pb_size, title=_title)
    else:
        pb = DummyProgressBar()
    groupmap = group_mapping(cnx, interactive=False)
    # serialize all entity types, assuring CWEType is serialized first for proper
    # is / is_instance_of insertion
    eschemas.remove(schema.entity_schema_for("CWEType"))
    eschemas.insert(0, schema.entity_schema_for("CWEType"))
    for eschema in eschemas:
        execschemarql(execute, eschema, eschema2rql(eschema, groupmap))
        pb.update()
    # serialize constraint types
    cstrtypemap = {}
    rql = "INSERT CWConstraintType X: X name %(ct)s"
    for cstrtype in CONSTRAINTS:
        cstrtypemap[cstrtype] = execute(rql, {"ct": cstrtype}, build_descr=False)[0][0]
        pb.update()
    # serialize relations
    for rschema in schema.relations():
        # skip virtual relations such as eid, has_text and identity
        if rschema in VIRTUAL_RTYPES:
            pb.update()
            continue
        if rschema.rule:
            execschemarql(execute, rschema, crschema2rql(rschema, groupmap))
            pb.update()
            continue
        execschemarql(execute, rschema, rschema2rql(rschema, addrdef=False))
        if rschema.symmetric:
            rdefs = [
                rdef
                for k, rdef in rschema.relation_definitions.items()
                if (rdef.subject, rdef.object) == k
            ]
        else:
            rdefs = rschema.relation_definitions.values()
        for rdef in rdefs:
            execschemarql(execute, rdef, rdef2rql(rdef, cstrtypemap, groupmap))
        pb.update()
    # serialize unique_together constraints
    for eschema in eschemas:
        if eschema._unique_together:
            execschemarql(execute, eschema, uniquetogether2rqls(eschema))
    # serialize yams inheritance relationships
    for rql, kwargs in specialize2rql(schema):
        execute(rql, kwargs, build_descr=False)
        pb.update()
    print()


# high level serialization functions


def execschemarql(execute, schema, rqls):
    for rql, kwargs in rqls:
        kwargs["x"] = schema.eid
        rset = execute(rql, kwargs, build_descr=False)
        if schema.eid is None:
            schema.eid = rset[0][0]
        else:
            assert rset


def erschema2rql(erschema, groupmap):
    if isinstance(erschema, schemamod.EntitySchema):
        return eschema2rql(erschema, groupmap=groupmap)
    return rschema2rql(erschema, groupmap=groupmap)


def specialize2rql(schema):
    for eschema in schema.entities():
        if eschema.final:
            continue
        for rql, kwargs in eschemaspecialize2rql(eschema):
            yield rql, kwargs


# etype serialization


def eschema2rql(eschema, groupmap=None):
    """return a list of rql insert statements to enter an entity schema
    in the database as an CWEType entity
    """
    relations, values = eschema_relations_values(eschema)
    # NOTE: 'specializes' relation can't be inserted here since there's no
    # way to make sure the parent type is inserted before the child type
    yield f"INSERT CWEType X: {','.join(relations)}", values
    # entity permissions
    if groupmap is not None:
        for rql, args in _erperms2rql(eschema, groupmap):
            yield rql, args


def eschema_relations_values(eschema):
    values = _ervalues(eschema)
    relations = [f"X {attr} %({attr})s" for attr in sorted(values)]
    return relations, values


def eschemaspecialize2rql(eschema):
    specialized_type = eschema.specializes()
    if specialized_type:
        values = {"x": eschema.eid, "et": specialized_type.eid}
        yield "SET X specializes ET WHERE X eid %(x)s, ET eid %(et)s", values


def uniquetogether2rqls(eschema):
    rql_args = []
    # robustness against duplicated CWUniqueTogetherConstraint (pre 3.18)
    columnset = set()
    for columns in eschema._unique_together:
        if columns in columnset:
            print(
                "schemaserial: skipping duplicate unique together %r %r"
                % (eschema.type, columns)
            )
            continue
        columnset.add(columns)
        rql, args = _uniquetogether2rql(eschema, columns)
        args["name"] = y2sql.unique_index_name(eschema, columns)
        rql_args.append((rql, args))
    return rql_args


def _uniquetogether2rql(eschema, unique_together):
    relations = []
    restrictions = []
    substs = {}
    for i, name in enumerate(unique_together):
        rschema = eschema.schema.relation_schema_for(name)
        rtype = "T%d" % i
        substs[rtype] = rschema.type
        relations.append(f"C relations {rtype}")
        restrictions.append("%(rtype)s name %%(%(rtype)s)s" % {"rtype": rtype})
    relations = ", ".join(relations)
    restrictions = ", ".join(restrictions)
    rql = (
        "INSERT CWUniqueTogetherConstraint C: C name %%(name)s, C constraint_of X, %s "
        "WHERE X eid %%(x)s, %s"
    )
    return rql % (relations, restrictions), substs


def _ervalues(erschema):
    return {
        "name": erschema.type,
        "final": erschema.final,
        "description": erschema.description,
    }


# rtype serialization


def rschema2rql(rschema, cstrtypemap=None, addrdef=True, groupmap=None):
    """generate rql insert statements to enter a relation schema
    in the database as an CWRType entity
    """
    if rschema.type == "has_text":
        return
    relations, values = rschema_relations_values(rschema)
    yield f"INSERT CWRType X: {','.join(relations)}", values
    if addrdef:
        assert cstrtypemap
        # sort for testing purpose
        for rdef in sorted(
            rschema.relation_definitions.values(), key=lambda x: (x.subject, x.object)
        ):
            for rql, values in rdef2rql(rdef, cstrtypemap, groupmap):
                yield rql, values


def rschema_relations_values(rschema):
    values = _ervalues(rschema)
    values["final"] = rschema.final
    values["symmetric"] = rschema.symmetric
    values["inlined"] = rschema.inlined
    values["fulltext_container"] = rschema.fulltext_container
    relations = [f"X {attr} %({attr})s" for attr in sorted(values)]
    return relations, values


def crschema2rql(crschema, groupmap):
    relations, values = crschema_relations_values(crschema)
    yield f"INSERT CWComputedRType X: {','.join(relations)}", values
    if groupmap:
        for rql, args in _erperms2rql(crschema, groupmap):
            yield rql, args


def crschema_relations_values(crschema):
    values = _ervalues(crschema)
    values["rule"] = crschema.rule
    # XXX why oh why?
    del values["final"]
    relations = [f"X {attr} %({attr})s" for attr in sorted(values)]
    return relations, values


# rdef serialization


def rdef2rql(rdef, cstrtypemap, groupmap=None):
    # don't serialize inferred relations
    if rdef.infered:
        return
    relations, values = _rdef_values(rdef)
    relations.append("X relation_type ER,X from_entity SE,X to_entity OE")
    values.update(
        {"se": rdef.subject.eid, "rt": rdef.relation_type.eid, "oe": rdef.object.eid}
    )
    if rdef.final:
        etype = "CWAttribute"
    else:
        etype = "CWRelation"
    yield "INSERT %s X: %s WHERE SE eid %%(se)s,ER eid %%(rt)s,OE eid %%(oe)s" % (
        etype,
        ",".join(relations),
    ), values
    for rql, values in constraints2rql(cstrtypemap, rdef.constraints):
        yield rql, values
    # no groupmap means "no security insertion"
    if groupmap:
        for rql, args in _erperms2rql(rdef, groupmap):
            yield rql, args


_IGNORED_PROPS = ["eid", "constraints", "uid", "infered", "permissions"]


def _rdef_values(rdef):
    amap = {"order": "ordernum", "default": "defaultval"}
    values = {}
    extra = {}
    for prop in rdef.rproperty_defs(rdef.object):
        if prop in _IGNORED_PROPS:
            continue
        value = getattr(rdef, prop)
        if prop not in KNOWN_RPROPERTIES:
            extra[prop] = value
            continue
        # XXX type cast really necessary?
        if prop in ("indexed", "fulltextindexed", "internationalizable"):
            value = bool(value)
        elif prop == "ordernum":
            value = int(value)
        if value is not None and prop == "default":
            value = Binary.zpickle(value)
        values[amap.get(prop, prop)] = value
    if extra:
        values["extra_props"] = Binary(json.dumps(extra).encode("ascii"))
    relations = [f"X {attr} %({attr})s" for attr in sorted(values)]
    return relations, values


def constraints2rql(cstrtypemap, constraints, rdefeid=None):
    for constraint in constraints:
        values = {
            "ct": cstrtypemap[constraint.type()],
            "value": constraint.serialize(),
            "x": rdefeid,
        }  # when not specified, will have to be set by the caller
        yield "INSERT CWConstraint X: X value %(value)s, X cstrtype CT, EDEF constrained_by X WHERE \
CT eid %(ct)s, EDEF eid %(x)s", values


def _erperms2rql(erschema, groupmap):
    """return rql insert statements to enter the entity or relation
    schema's permissions in the database as
    [read|add|delete|update]_permission relations between CWEType/CWRType
    and CWGroup entities
    """
    for action in erschema.ACTIONS:
        try:
            grantedto = erschema.action_permissions(action)
        except KeyError:
            # may occurs when modifying persistent schema
            continue
        for group_or_rqlexpr in grantedto:
            if isinstance(group_or_rqlexpr, str):
                # group
                try:
                    yield (
                        "SET X %s_permission Y WHERE Y eid %%(g)s, X eid %%(x)s"
                        % action,
                        {"g": groupmap[group_or_rqlexpr]},
                    )
                except KeyError:
                    print(
                        "WARNING: group %s used in permissions for %s was ignored because it doesn't exist."
                        " You may want to add it into a precreate.py file"
                        % (group_or_rqlexpr, erschema)
                    )
                    continue
            else:
                # rqlexpr
                rqlexpr = group_or_rqlexpr
                yield (
                    "INSERT RQLExpression E: E expression %%(e)s, E exprtype %%(t)s, "
                    "E mainvars %%(v)s, X %s_permission E WHERE X eid %%(x)s" % action,
                    {
                        "e": rqlexpr.expression,
                        "v": ",".join(sorted(rqlexpr.mainvars)),
                        "t": rqlexpr.__class__.__name__,
                    },
                )


# update functions


def updateeschema2rql(eschema, eid):
    relations, values = eschema_relations_values(eschema)
    values["x"] = eid
    yield "SET %s WHERE X eid %%(x)s" % ",".join(relations), values


def updaterschema2rql(rschema, eid):
    if rschema.rule:
        yield ("SET X rule %(r)s WHERE X eid %(x)s", {"x": eid, "r": rschema.rule})
    else:
        relations, values = rschema_relations_values(rschema)
        values["x"] = eid
        yield "SET %s WHERE X eid %%(x)s" % ",".join(relations), values


def updaterdef2rql(rdef, eid):
    relations, values = _rdef_values(rdef)
    values["x"] = eid
    yield "SET %s WHERE X eid %%(x)s" % ",".join(relations), values



# copyright 2003-2015 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact https://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of CubicWeb.
#
# CubicWeb is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# CubicWeb is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with CubicWeb.  If not, see <https://www.gnu.org/licenses/>.
"""unittest for cubicweb.dataimport.csv"""

from io import BytesIO

from logilab.common.testlib import TestCase, unittest_main

from cubicweb.dataimport import csv


class UcsvreaderTC(TestCase):
    def test_empty_lines_skipped(self):
        stream = BytesIO(
            b"""a,b,c,d,
1,2,3,4,
,,,,
,,,,
"""
        )
        self.assertEqual(
            [
                ["a", "b", "c", "d", ""],
                ["1", "2", "3", "4", ""],
            ],
            list(csv.ucsvreader(stream)),
        )
        stream.seek(0)
        self.assertEqual(
            [
                ["a", "b", "c", "d", ""],
                ["1", "2", "3", "4", ""],
                ["", "", "", "", ""],
                ["", "", "", "", ""],
            ],
            list(csv.ucsvreader(stream, skip_empty=False)),
        )

    def test_skip_first(self):
        stream = BytesIO(b"a,b,c,d,\n1,2,3,4,\n")
        reader = csv.ucsvreader(stream, skipfirst=True, ignore_errors=True)
        self.assertEqual(list(reader), [["1", "2", "3", "4", ""]])

        stream.seek(0)
        reader = csv.ucsvreader(stream, skipfirst=True, ignore_errors=False)
        self.assertEqual(list(reader), [["1", "2", "3", "4", ""]])

        stream.seek(0)
        reader = csv.ucsvreader(stream, skipfirst=False, ignore_errors=True)
        self.assertEqual(
            list(reader), [["a", "b", "c", "d", ""], ["1", "2", "3", "4", ""]]
        )

        stream.seek(0)
        reader = csv.ucsvreader(stream, skipfirst=False, ignore_errors=False)
        self.assertEqual(
            list(reader), [["a", "b", "c", "d", ""], ["1", "2", "3", "4", ""]]
        )


if __name__ == "__main__":
    unittest_main()



# copyright 2003-2016 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact https://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of CubicWeb.
#
# CubicWeb is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# CubicWeb is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with CubicWeb.  If not, see <https://www.gnu.org/licenses/>.
"""data source related entities"""

import logging
import re
from socket import gethostname

from logilab.common.textutils import text_to_dict
from logilab.mtconverter import xml_escape

from cubicweb.entities import AnyEntity, fetch_config


class _CWSourceCfgMixIn:
    @property
    def dictconfig(self):
        if not self.config:
            return {}
        return text_to_dict(self.config)


class CWSource(_CWSourceCfgMixIn, AnyEntity):
    __regid__ = "CWSource"
    fetch_attrs, cw_fetch_order = fetch_config(["name", "type"])

    @property
    def host_config(self):
        dictconfig = self.dictconfig
        host = gethostname()
        for hostcfg in self.host_configs:
            if hostcfg.match(host):
                self.info(
                    "matching host config %s for source %s",
                    hostcfg.match_host,
                    self.name,
                )
                dictconfig.update(hostcfg.dictconfig)
        return dictconfig

    @property
    def host_configs(self):
        return self.reverse_cw_host_config_of

    @property
    def repo_source(self):
        """repository only property, not available from the web side (eg
        self._cw is expected to be a server session)
        """
        return self._cw.repo.source_by_eid(self.eid)


class CWSourceHostConfig(_CWSourceCfgMixIn, AnyEntity):
    __regid__ = "CWSourceHostConfig"
    fetch_attrs, cw_fetch_order = fetch_config(["match_host", "config"])

    @property
    def cwsource(self):
        return self.cw_host_config_of[0]

    def match(self, hostname):
        return re.match(self.match_host, hostname)


class CWDataImport(AnyEntity):
    __regid__ = "CWDataImport"
    repo_source = _logs = None  # please pylint

    def init(self):
        self._logs = []
        self.repo_source = self.cwsource.repo_source

    def dc_title(self):
        return "{} [{}]".format(
            self.printable_value("start_timestamp"),
            self.printable_value("status"),
        )

    @property
    def cwsource(self):
        return self.cw_import_of[0]

    def record_debug(self, msg, path=None, line=None):
        self._log(logging.DEBUG, msg, path, line)
        self.repo_source.debug(msg)

    def record_info(self, msg, path=None, line=None):
        self._log(logging.INFO, msg, path, line)
        self.repo_source.info(msg)

    def record_warning(self, msg, path=None, line=None):
        self._log(logging.WARNING, msg, path, line)
        self.repo_source.warning(msg)

    def record_error(self, msg, path=None, line=None):
        self._status = "failed"
        self._log(logging.ERROR, msg, path, line)
        self.repo_source.error(msg)

    def record_fatal(self, msg, path=None, line=None):
        self._status = "failed"
        self._log(logging.FATAL, msg, path, line)
        self.repo_source.fatal(msg)

    def _log(self, severity, msg, path=None, line=None):
        encodedmsg = f"{severity}\t{path or ''}\t{line or ''}\t{xml_escape(msg)}<br/>"
        self._logs.append(encodedmsg)

    def write_log(self, session, **kwargs):
        if "status" not in kwargs:
            kwargs["status"] = getattr(self, "_status", "success")
        self.cw_set(log="<br/>".join(self._logs), **kwargs)
        self._logs = []



# copyright 2013-2014 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact https://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of CubicWeb.
#
# CubicWeb is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# CubicWeb is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with CubicWeb.  If not, see <https://www.gnu.org/licenses/>.
"""Official API to access the content of a repository."""

from cubicweb import AuthenticationError
from cubicweb.server.session import Connection


def get_repository(config, vreg=None):
    """get a repository for the given URI or config/vregistry (in case we're
    loading the repository for a client, eg web server, configuration).

    The returned repository may be an in-memory repository or a proxy object
    using a specific RPC method, depending on the given URI.
    """
    return config.repository(vreg)


def connect(repo, login, **kwargs):
    """Take credential and return associated Connection.

    raise AuthenticationError if the credential are invalid.
    """
    # use an internal connection to try to get a user object
    with repo.internal_cnx() as cnx:
        user = repo.authenticate_user(cnx, login, **kwargs)
    return Connection(repo, user)


def anonymous_cnx(repo):
    """return a Connection for Anonymous user.

    raises an AuthenticationError if anonymous usage is not allowed
    """
    anoninfo = getattr(repo.config, "anonymous_user", lambda: None)()
    if anoninfo is None:  # no anonymous user
        raise AuthenticationError("anonymous access is not authorized")
    anon_login, anon_password = anoninfo
    # use vreg's repository cache
    return connect(repo, anon_login, password=anon_password)



# copyright 2017-2025 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# copyright 2014-2016 UNLISH S.A.S. (Montpellier, FRANCE), all rights reserved.
#
# contact https://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of CubicWeb.
#
# CubicWeb is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# CubicWeb is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with CubicWeb.  If not, see <https://www.gnu.org/licenses/>.

"""Pyramid interface to CubicWeb"""

import atexit
import logging
import os
import sys
from configparser import ConfigParser
from urllib.parse import urlparse

import wsgicors
from pyramid.config import Configurator
from pyramid.exceptions import ConfigurationError
from pyramid.settings import asbool, aslist

from cubicweb.cwconfig import CubicWebConfiguration as cwcfg
from cubicweb.pyramid.csrf import CWCookieCSRFStoragePolicy


def set_csrf_storage_policy(config, base_url) -> None:
    # don't inline it, so it can be easily monkeypatched
    # see https://forge.extranet.logilab.fr/cubicweb/cubicweb/-/issues/1080
    config.set_csrf_storage_policy(
        CWCookieCSRFStoragePolicy(path=urlparse(base_url).path)
    )


def config_from_cwconfig(cwconfig, settings=None, debugtoolbar=False):
    """Return a Pyramid Configurator instance built from a CubicWeb config and
    Pyramid-specific configuration files (pyramid.ini).

    :param cwconfig: A CubicWeb configuration
    :returns: A Pyramid config object
    """
    settings = dict(settings) if settings else {}
    settings.update(settings_from_cwconfig(cwconfig, debugtoolbar=debugtoolbar))
    config = Configurator(settings=settings)
    config.registry["cubicweb.config"] = cwconfig
    config.include("cubicweb.pyramid")

    if debugtoolbar:
        config.include("pyramid_debugtoolbar")

    return config


def settings_from_cwconfig(cwconfig, debugtoolbar=False) -> dict:
    """
    Extract settings from pyramid.ini and pyramid-debug.ini (if in debug)

    Can be used to configure middleware WSGI with settings from pyramid.ini files

    :param cwconfig: A CubicWeb configuration
    :returns: A settings dictionnary
    """
    settings_filenames = [os.path.join(cwconfig.apphome, "pyramid.ini")]
    settings = {}
    if cwconfig.debugmode:
        settings_filenames.insert(
            0, os.path.join(cwconfig.apphome, "pyramid-debug.ini")
        )

        settings.update(
            {
                "pyramid.debug_authorization": True,
                "pyramid.debug_notfound": True,
                "pyramid.debug_routematch": True,
                "pyramid.reload_templates": True,
            }
        )

        if debugtoolbar:
            settings["debugtoolbar.includes"] = ["cubicweb.pyramid.debugtoolbar_panels"]

    for fname in settings_filenames:
        if os.path.exists(fname):
            cp = ConfigParser()
            cp.read(fname)
            settings.update(cp.items("main"))
            break

    return settings


def wsgi_application_from_cwconfig(
    cwconfig,
    debugtoolbar=False,
):
    """Build a WSGI application from a cubicweb configuration

    :param cwconfig: A CubicWeb configuration
    :param profile: Enable profiling. See :ref:`profiling`.
    :param profile_output: Profiling output filename. See :ref:`profiling`.
    :param profile_dump_every: Profiling number of requests before dumping the
                               stats. See :ref:`profiling`.
    :param debugtoolbar: Activate pyramid debugtoolbar when True.

    :returns: A fully operationnal WSGI application
    """
    config = config_from_cwconfig(cwconfig, debugtoolbar=debugtoolbar)

    if debugtoolbar:
        try:
            from cubicweb.pyramid.debug_source_code import (
                debug_display_source_code,
                DEBUG_DISPLAY_SOURCE_CODE_PATH,
            )
        except ImportError as e:
            logging.exception(
                "Error: you tried to activate the debug toolbar but you have missing "
                "dependencies: %s" % e
            )

            sys.exit(1)
        else:
            config.add_route(
                "debug_display_source_code", DEBUG_DISPLAY_SOURCE_CODE_PATH
            )
            config.add_view(
                debug_display_source_code, route_name="debug_display_source_code"
            )

    app = config.make_wsgi_app()
    app = wsgicors.CORS(
        app,
        origin=" ".join(cwconfig["access-control-allow-origin"]),
        headers=", ".join(cwconfig["access-control-allow-headers"]),
        expose_headers=", ".join(cwconfig["access-control-expose-headers"]),
        methods=", ".join(cwconfig["access-control-allow-methods"]),
        maxage=str(cwconfig["access-control-max-age"]),
        credentials="true",
    )
    return app


def wsgi_application(instance_name=None, debug=None):
    """Build a WSGI application from a cubicweb instance name

    :param instance_name: Name of the cubicweb instance (optional). If not
                          provided, :envvar:`CW_INSTANCE` must exists.
    :param debug: Enable/disable the debug mode. If defined to True or False,
                  overrides :envvar:`CW_DEBUG`.

    The following environment variables are used if they exist:

    .. envvar:: CW_INSTANCE

        A CubicWeb instance name.

    .. envvar:: CW_DEBUG

        If defined, the debugmode is enabled.

    The function can be used as an entry-point for third-party wsgi containers.
    Below is a sample uswgi configuration file:

    .. code-block:: ini

        [uwsgi]
        http = 127.0.1.1:8080
        env = CW_INSTANCE=myinstance
        env = CW_DEBUG=1
        module = cubicweb.pyramid:wsgi_application()
        virtualenv = /home/user/.virtualenvs/myvirtualenv
        processes = 1
        threads = 8
        stats = 127.0.0.1:9191
        plugins = http,python

    """
    if instance_name is None:
        instance_name = os.environ["CW_INSTANCE"]
    if debug is None:
        debug = "CW_DEBUG" in os.environ

    cwconfig = cwcfg.config_for(instance_name, debugmode=debug)

    return wsgi_application_from_cwconfig(cwconfig)


def pyramid_app(global_config, **settings):
    """Return a Pyramid WSGI application bound to a CubicWeb repository."""
    config = Configurator(settings=settings)
    config.include("cubicweb.pyramid")
    return config.make_wsgi_app()


def includeme(config) -> None:
    """Set-up a CubicWeb instance.

    The CubicWeb instance can be set in several ways:

    -   Provide an already loaded CubicWeb repository in the registry:

        .. code-block:: python

            config.registry['cubicweb.repository'] = your_repo_instance

    -   Provide an already loaded CubicWeb config instance in the registry:

        .. code-block:: python

            config.registry['cubicweb.config'] = your_config_instance

    -   Provide an instance name in the pyramid settings with
        :confval:`cubicweb.instance`.

    A CubicWeb repository is instantiated and attached in
    'cubicweb.repository' registry key if not already present.

    The CubicWeb instance registry is attached in 'cubicweb.registry' registry
    key.
    """
    cwconfig = config.registry.get("cubicweb.config")
    repo = config.registry.get("cubicweb.repository")

    if repo is not None:
        if cwconfig is None:
            config.registry["cubicweb.config"] = cwconfig = repo.config
        elif cwconfig is not repo.config:
            raise ConfigurationError(
                'CubicWeb config instance (found in "cubicweb.config" '
                "registry key) mismatches with that of the repository "
                '(registry["cubicweb.repository"])'
            )

    if cwconfig is None:
        debugmode = asbool(config.registry.settings.get("cubicweb.debug", False))
        cwconfig = cwcfg.config_for(
            config.registry.settings["cubicweb.instance"], debugmode=debugmode
        )
        config.registry["cubicweb.config"] = cwconfig

    if repo is None:
        repo = config.registry["cubicweb.repository"] = cwconfig.repository()
    config.registry["cubicweb.registry"] = repo.vreg

    base_url = cwconfig["base-url"]
    assert base_url is not None, (
        "base-url isn't properly set, it can be an issue in CubicWeb source code or in your configuration."
        "In normal case, 'base-url' is defined by cwconfig.repository()."
    )
    if cwconfig["receives-base-url-path"]:
        config.route_prefix = urlparse(base_url).path

    if cwconfig.mode != "test":

        @atexit.register
        def shutdown_repo():
            if repo.shutting_down:
                return
            repo.shutdown

    config.include("cubicweb.pyramid.predicates")

    cubicweb_includes = aslist(config.registry.settings.get("cubicweb.includes", []))
    if (
        "cubicweb.pyramid.rest_api.include_html" in cubicweb_includes
        or "cubicweb.pyramid.rest_api.include_rdf" in cubicweb_includes
    ):
        config.include("cubicweb.pyramid.rest_api")

    for name in cubicweb_includes:
        config.include(name)

    config.include("cubicweb.pyramid.core")

    config.set_default_csrf_options(
        require_csrf=asbool(
            config.registry.settings.get("cubicweb.pyramid.enable_csrf", "yes")
        )
    )
    set_csrf_storage_policy(config, base_url)



# copyright 2003-2015 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact https://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of CubicWeb.
#
# CubicWeb is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# CubicWeb is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with CubicWeb.  If not, see <https://www.gnu.org/licenses/>.
"""unittest for cubicweb.dataimport.pgstore"""

import datetime as DT

from logilab.common.testlib import TestCase, unittest_main

from cubicweb.dataimport import pgstore


class CreateCopyFromBufferTC(TestCase):
    # test converters

    def test_convert_none(self):
        cnvt = pgstore._copyfrom_buffer_convert_None
        self.assertEqual("NULL", cnvt(None))

    def test_convert_number(self):
        cnvt = pgstore._copyfrom_buffer_convert_number
        self.assertEqual("42", cnvt(42))
        self.assertEqual("42", cnvt(int(42)))
        self.assertEqual("42.42", cnvt(42.42))

    def test_convert_string(self):
        cnvt = pgstore._copyfrom_buffer_convert_string
        # simple
        self.assertEqual("babar", cnvt("babar"))
        # unicode
        self.assertEqual("éléphant", cnvt("éléphant"))
        # escaping
        self.assertEqual("babar\\tceleste\\n", cnvt("babar\tceleste\n"))
        self.assertEqual("C:\\\\new\\tC:\\\\test", cnvt("C:\\new\tC:\\test"))

    def test_convert_date(self):
        cnvt = pgstore._copyfrom_buffer_convert_date
        self.assertEqual("0666-01-13", cnvt(DT.date(666, 1, 13)))

    def test_convert_time(self):
        cnvt = pgstore._copyfrom_buffer_convert_time
        self.assertEqual("06:06:06.000100", cnvt(DT.time(6, 6, 6, 100)))

    def test_convert_datetime(self):
        cnvt = pgstore._copyfrom_buffer_convert_datetime
        self.assertEqual(
            "0666-06-13 06:06:06.000000", cnvt(DT.datetime(666, 6, 13, 6, 6, 6))
        )

    # test buffer
    def test_create_copyfrom_buffer_tuple(self):
        data = (
            (
                42,
                int(42),
                42.42,
                "éléphant",
                DT.date(666, 1, 13),
                DT.time(6, 6, 6),
                DT.datetime(666, 6, 13, 6, 6, 6),
            ),
            (
                6,
                int(6),
                6.6,
                "babar",
                DT.date(2014, 1, 14),
                DT.time(4, 2, 1),
                DT.datetime(2014, 1, 1, 0, 0, 0),
            ),
        )
        results = pgstore._create_copyfrom_buffer(data)
        # all columns
        expected = """42\t42\t42.42\téléphant\t0666-01-13\t06:06:06.000000\t0666-06-13 06:06:06.000000
6\t6\t6.6\tbabar\t2014-01-14\t04:02:01.000000\t2014-01-01 00:00:00.000000"""
        self.assertMultiLineEqual(expected, results.getvalue())
        # selected columns
        results = pgstore._create_copyfrom_buffer(data, columns=(1, 3, 6))
        expected = """42\téléphant\t0666-06-13 06:06:06.000000
6\tbabar\t2014-01-01 00:00:00.000000"""
        self.assertMultiLineEqual(expected, results.getvalue())

    def test_create_copyfrom_buffer_dict(self):
        data = (
            dict(
                integer=42,
                double=42.42,
                text="éléphant",
                date=DT.datetime(666, 6, 13, 6, 6, 6),
            ),
            dict(
                integer=6,
                double=6.6,
                text="babar",
                date=DT.datetime(2014, 1, 1, 0, 0, 0),
            ),
        )
        results = pgstore._create_copyfrom_buffer(data, ("integer", "text"))
        expected = """42\téléphant\n6\tbabar"""
        self.assertEqual(expected, results.getvalue())


if __name__ == "__main__":
    unittest_main()



# copyright 2003-2023 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact https://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of CubicWeb.
#
# CubicWeb is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# CubicWeb is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with CubicWeb.  If not, see <https://www.gnu.org/licenses/>.
"""Base class for request/session"""

from datetime import time, datetime, timedelta
from urllib.parse import (
    parse_qs,
    parse_qsl,
    quote as urlquote,
    unquote as urlunquote,
    urlsplit,
    urlunsplit,
)

from logilab.common.date import ustrftime, todate, todatetime
from logilab.common.decorators import cached
from rql.utils import rqlvar_maker

from cubicweb import REVERSE_INDICATOR, Unauthorized, NoSelectableObject, uilib
from cubicweb.rset import ResultSet

ONESECOND = timedelta(0, 1, 0)
CACHE_REGISTRY = {}


class Cache(dict):
    def __init__(self):
        super(Cache, self).__init__()
        _now = datetime.now()
        self.cache_creation_date = _now
        self.latest_cache_lookup = _now


class RequestSessionAndConnectionBase:
    """base class containing stuff shared by server session and web request

    request/session is the main resources accessor, mainly through it's vreg
    attribute:

    :attribute vreg: the instance's registry
    :attribute vreg.schema: the instance's schema
    :attribute vreg.config: the instance's configuration
    """

    is_request = True  # False for repository session

    def __init__(self, vreg):
        self.vreg = vreg
        try:
            encoding = vreg.property_value("ui.encoding")
        except Exception:  # no vreg or property not registered
            encoding = "utf-8"
        self.encoding = encoding
        # cache result of execution for (rql expr / eids),
        # should be emptied on commit/rollback of the server session / web
        # connection
        self.user = None
        self.lang = None
        self.local_perm_cache = {}
        self._ = str

    def _set_user(self, orig_user):
        """set the user for this req_session_base

        A special method is needed to ensure the linked user is linked to the
        connection too.
        """
        rset = self.eid_rset(orig_user.eid, "CWUser")
        user = self.vreg["etypes"].etype_class("CWUser")(self, rset, row=0)
        user.cw_attr_cache["login"] = orig_user.login  # cache login
        self.user = user
        self.set_entity_cache(user)

    def set_language(self, lang):
        """install i18n configuration for `lang` translation.

        Raises :exc:`KeyError` if translation doesn't exist.
        """
        self.lang = lang
        try:
            gettext, pgettext = self.vreg.config.translations[lang]
        except KeyError:
            assert self.vreg.config.mode == "test"
            gettext = str

            def pgettext(x, y):
                return str(y)

        # use _cw.__ to translate a message without registering it to the catalog
        self._ = self.__ = gettext
        self.pgettext = pgettext

    def get_option_value(self, option):
        raise NotImplementedError

    def property_value(self, key):
        """return value of the property with the given key, giving priority to
        user specific value if any, else using site value
        """
        if self.user:
            val = self.user.property_value(key)
            if val is not None:
                return val
        return self.vreg.property_value(key)

    def etype_rset(self, etype, size=1):
        """return a fake result set for a particular entity type"""
        rset = ResultSet([("A",)] * size, f"{etype} X", description=[(etype,)] * size)

        def get_entity(row, col=0, etype=etype, req=self, rset=rset):
            return req.vreg["etypes"].etype_class(etype)(req, rset, row, col)

        rset.get_entity = get_entity
        rset.req = self
        return rset

    def eid_rset(self, eid, etype=None):
        """return a result set for the given eid without doing actual query
        (we have the eid, we can suppose it exists and user has access to the
        entity)
        """
        eid = int(eid)
        if etype is None:
            etype = self.entity_type(eid)
        rset = ResultSet([(eid,)], "Any X WHERE X eid %(x)s", {"x": eid}, [(etype,)])
        rset.req = self
        return rset

    def empty_rset(self):
        """return a guaranteed empty result"""
        rset = ResultSet([], "Any X WHERE X eid -1")
        rset.req = self
        return rset

    def entity_from_eid(self, eid, etype=None):
        """return an entity instance for the given eid. No query is done"""
        try:
            return self.entity_cache(eid)
        except KeyError:
            rset = self.eid_rset(eid, etype)
            entity = rset.get_entity(0, 0)
            self.set_entity_cache(entity)
            return entity

    def entity_cache(self, eid):
        raise KeyError

    def set_entity_cache(self, entity):
        pass

    def create_entity(self, etype, **kwargs):
        """add a new entity of the given type

        Example (in a shell session):

        >>> c = req.create_entity('Company', name='Logilab')
        >>> req.create_entity('Person', firstname='John', surname='Doe',
        ...                   works_for=c)

        """
        cls = self.vreg["etypes"].etype_class(etype)
        return cls.cw_instantiate(self.execute, **kwargs)

    def _build_rql_request(self, start, etype, **kwargs):
        parts = [f"{start:s} WHERE X is {etype:s}"]
        varmaker = rqlvar_maker(defined="X")
        eschema = self.vreg.schema.entity_schema_for(etype)
        for attr, value in kwargs.items():
            if isinstance(value, (list, tuple)):
                raise NotImplementedError(
                    f"{attr}: collection of values is not supported"
                )
            if hasattr(value, "eid"):
                kwargs[attr] = value.eid
            if attr.startswith(REVERSE_INDICATOR):
                attr = attr[len(REVERSE_INDICATOR) :]
                if attr not in eschema.object_relations:
                    raise KeyError(f"{attr:s} not in {eschema} object relations")
                parts.append(
                    "{var} {attr} X, {var} eid %({reverse_ind:s}{attr})s"
                    "".format(
                        var=next(varmaker),
                        attr=attr,
                        reverse_ind=REVERSE_INDICATOR,
                    )
                )
            else:
                rel = eschema.subject_relations.get(attr)
                if rel is None:
                    raise KeyError(f"{attr} not in {eschema} subject relations")
                if rel.final:
                    parts.append("X {attr} %({attr})s".format(attr=attr))
                else:
                    parts.append(
                        "X {attr} {var}, {var} eid %({attr})s".format(
                            attr=attr, var=next(varmaker)
                        )
                    )

        rql = ", ".join(parts)
        return rql, kwargs

    def exists(self, etype, **kwargs):
        """return if it exists at least one entity
        of the given type and attribute values.

        >>> is_stallman_there = cnx.exists('CWUser', login=u"rms")
        >>> # more optimized than
        >>> is_stallman_there = bool(cnx.find('CWUser', login=u"rms"))
        >>> # it did a full scan table :-/
        """
        try:
            del kwargs["limit"]
        except KeyError:
            pass
        return bool(self.find(etype, limit=1, **kwargs))

    def find(self, etype, limit=None, **kwargs):
        """find entities of the given type and attribute values.

        :param etype: the type of entities to return.
        :param limit: the max number of entities to return.
        :returns: A :class:`ResultSet`

        >>> users = find('CWGroup', name=u"users").one()
        >>> groups = find('CWGroup').entities()
        """
        base_rql = "Any X"
        if limit is not None:
            base_rql += f" LIMIT {limit:d}"
        rql, kwargs = self._build_rql_request(base_rql, etype, **kwargs)
        return self.execute(rql, kwargs)

    def ensure_ro_rql(self, rql):
        """raise an exception if the given rql is not a select query"""
        first = rql.split(None, 1)[0].lower()
        if first in ("insert", "set", "delete"):
            raise Unauthorized(self._("only select queries are authorized"))

    # url generation methods ##################################################

    def build_url(self, *args, **kwargs):
        """return an absolute URL using params dictionary key/values as URL
        parameters. Values are automatically URL quoted, and the
        publishing method to use may be specified or will be guessed.

        raises :exc:`ValueError` if None is found in arguments
        """
        # use *args since we don't want first argument to be "anonymous" to
        # avoid potential clash with kwargs
        method = None
        if args:
            assert len(args) == 1, "only 0 or 1 non-named-argument expected"
            method = args[0]
        if method is None:
            method = "view"
        # XXX I (adim) think that if method is passed explicitly, we should
        #     not try to process it and directly call req.build_url()
        base_url = kwargs.pop("base_url", None)
        if base_url is None:
            base_url = self.base_url()
        path = self.build_url_path(method, kwargs)
        if not kwargs:
            return f"{base_url}{path}"
        return f"{base_url}{path}?{self.build_url_params(**kwargs)}"

    def build_url_path(self, method, kwargs):
        """return the "path" part of an URL"""
        if "_restpath" in kwargs:
            assert method == "view", repr(method)
            path = kwargs.pop("_restpath")
        else:
            path = method
        return path

    def build_url_params(self, **kwargs):
        """return encoded params to incorporate them in a URL"""
        args = []
        for param, values in kwargs.items():
            if not isinstance(values, (list, tuple)):
                values = (values,)
            for value in values:
                assert value is not None
                args.append(f"{param}={self.url_quote(value)}")
        return "&".join(args)

    def url_quote(self, value, safe=""):
        """urllib.quote is not unicode safe, use this method to do the
        necessary encoding / decoding. Also it's designed to quote each
        part of a url path and so the '/' character will be encoded as well.
        """
        return urlquote(str(value), safe=safe)

    def url_unquote(self, quoted):
        """returns a unicode unquoted string

        decoding is based on `self.encoding` which is the encoding
        used in `url_quote`
        """
        return urlunquote(quoted)

    def url_parse_qsl(self, querystring):
        """return a list of (key, val) found in the url quoted query string"""
        for key, val in parse_qsl(querystring):
            yield key, val
        return

    def rebuild_url(self, url, **newparams):
        """return the given url with newparams inserted. If any new params
        is already specified in the url, it's overriden by the new value

        newparams may only be mono-valued.
        """
        schema, netloc, path, query, fragment = urlsplit(url)
        query = parse_qs(query)
        # sort for testing predictability
        for key, val in sorted(newparams.items()):
            query[key] = (self.url_quote(val),)
        query = "&".join(
            f"{param}={value}"
            for param, values in sorted(query.items())
            for value in values
        )
        return urlunsplit((schema, netloc, path, query, fragment))

    # bound user related methods ###############################################

    @cached
    def user_data(self):
        """returns a dictionary with this user's information.

        The keys are :

        login
            The user login

        name
            The user name, returned by user.name()

        email
            The user principal email

        """
        userinfo = {}
        user = self.user
        userinfo["login"] = user.login
        userinfo["name"] = user.name()
        userinfo["email"] = user.cw_adapt_to("IEmailable").get_email()
        return userinfo

    # formating methods #######################################################

    def view(
        self,
        __vid,
        rset=None,
        __fallback_oid=None,
        __registry="views",
        initargs=None,
        w=None,
        **kwargs,
    ):
        """Select object with the given id (`__oid`) then render it.  If the
        object isn't selectable, try to select fallback object if
        `__fallback_oid` is specified.

        If specified `initargs` is expected to be a dictionary containing
        arguments that should be given to selection (hence to object's __init__
        as well), but not to render(). Other arbitrary keyword arguments will be
        given to selection *and* to render(), and so should be handled by
        object's call or cell_call method..
        """
        if initargs is None:
            initargs = kwargs
        else:
            initargs.update(kwargs)
        try:
            view = self.vreg[__registry].select(__vid, self, rset=rset, **initargs)
        except NoSelectableObject:
            if __fallback_oid is None:
                raise
            view = self.vreg[__registry].select(
                __fallback_oid, self, rset=rset, **initargs
            )
        return view.render(w=w, **kwargs)

    def printable_value(
        self, attrtype, value, props=None, displaytime=True, formatters=uilib.PRINTERS
    ):
        """return a displayablye value (i.e. unicode string)"""
        if value is None:
            return ""
        try:
            as_string = formatters[attrtype]
        except KeyError:
            self.error("given bad attrtype %s", attrtype)
            return str(value)
        return as_string(value, self, props, displaytime)

    def format_date(self, date, date_format=None, time=False):
        """return a string for a date time according to instance's
        configuration
        """
        if date is not None:
            if date_format is None:
                if time:
                    date_format = self.property_value("ui.datetime-format")
                else:
                    date_format = self.property_value("ui.date-format")
            return ustrftime(date, date_format)
        return ""

    def format_time(self, time):
        """return a string for a time according to instance's
        configuration
        """
        if time is not None:
            return ustrftime(time, self.property_value("ui.time-format"))
        return ""

    def format_float(self, num):
        """return a string for floating point number according to instance's
        configuration
        """
        if num is not None:
            return self.property_value("ui.float-format") % num
        return ""

    def parse_datetime(self, value, etype="Datetime"):
        """get a datetime or time from a string (according to etype)
        Datetime formatted as Date are accepted
        """
        assert etype in ("Datetime", "Date", "Time"), etype
        # XXX raise proper validation error
        if etype == "Datetime":
            format = self.property_value("ui.datetime-format")
            try:
                return todatetime(datetime.strptime(value, format))
            except ValueError:
                pass
        elif etype == "Time":
            format = self.property_value("ui.time-format")
            try:
                # (adim) I can't find a way to parse a time with a custom format
                date = datetime.strptime(value, format)  # this returns a datetime
                return time(date.hour, date.minute, date.second)
            except ValueError:
                raise ValueError(
                    self._("can't parse %(value)r (expected %(format)s)")
                    % {"value": value, "format": format}
                )
        try:
            format = self.property_value("ui.date-format")
            dt = datetime.strptime(value, format)
            if etype == "Datetime":
                return todatetime(dt)
            return todate(dt)
        except ValueError:
            raise ValueError(
                self._("can't parse %(value)r (expected %(format)s)")
                % {"value": value, "format": format}
            )

    def base_url(self):
        """Return the root url of the instance."""
        url = self.vreg.config["base-url"]
        return url if url is None else url.rstrip("/") + "/"



# copyright 2015-2016 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact https://www.logilab.fr -- mailto:contact@logilab.fr
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with this program. If not, see <https://www.gnu.org/licenses/>.
"""Tests for cubicweb.dataimport.importer"""

from logilab.common.testlib import TestCase, unittest_main

from yams import ValidationError
from cubicweb import Binary
from cubicweb.dataimport.stores import RQLObjectStore
from cubicweb.dataimport.csv import ucsvreader
from cubicweb.dataimport.importer import (
    ExtEntity,
    ExtEntitiesImporter,
    RelationMapping,
    SimpleImportLog,
    use_extid_as_cwuri,
    drop_extra_values,
)
from cubicweb.devtools.testlib import CubicWebTC


class RelationMappingTC(CubicWebTC):
    def test_nosource(self):
        with self.admin_access.repo_cnx() as cnx:
            alice_eid = cnx.create_entity("Personne", nom="alice").eid
            bob_eid = cnx.create_entity("Personne", nom="bob", connait=alice_eid).eid
            cnx.commit()
            mapping = RelationMapping(cnx)
            self.assertEqual(
                mapping["connait"], {(bob_eid, alice_eid), (alice_eid, bob_eid)}
            )

    def test_with_source(self):
        with self.admin_access.repo_cnx() as cnx:
            alice_eid = cnx.create_entity("Personne", nom="alice").eid
            bob_eid = cnx.create_entity("Personne", nom="bob", connait=alice_eid).eid
            cnx.commit()
            mapping = RelationMapping(cnx, cnx.find("CWSource", name="system").one())
            self.assertEqual(
                mapping["connait"], {(bob_eid, alice_eid), (alice_eid, bob_eid)}
            )


class ExtEntitiesImporterTC(CubicWebTC):
    def importer(self, cnx):
        store = RQLObjectStore(cnx)
        return ExtEntitiesImporter(self.schema, store, raise_on_error=True)

    def test_simple_import(self):
        with self.admin_access.repo_cnx() as cnx:
            importer = self.importer(cnx)
            personne = ExtEntity(
                "Personne", 1, {"nom": {"de la lune"}, "prenom": {"Jean"}}
            )
            importer.import_entities([personne])
            cnx.commit()
            rset = cnx.execute("Any X WHERE X is Personne")
            entity = rset.get_entity(0, 0)
            self.assertEqual(entity.nom, "de la lune")
            self.assertEqual(entity.prenom, "Jean")

    def test_bytes_attribute(self):
        with self.admin_access.repo_cnx() as cnx:
            importer = self.importer(cnx)
            personne = ExtEntity("Personne", 1, {"photo": {b"poilu"}, "nom": {"alf"}})
            importer.import_entities([personne])
            cnx.commit()
            entity = cnx.find("Personne").one()
            self.assertEqual(entity.photo.getvalue(), b"poilu")

    def test_binary_in_values(self):
        with self.admin_access.repo_cnx() as cnx:
            importer = self.importer(cnx)
            # Use a list to put a Binary in "values" (since Binary is not
            # hashable, a set cannot be used).
            personne = ExtEntity(
                "Personne", 1, {"photo": [Binary(b"poilu")], "nom": {"alf"}}
            )
            importer.import_entities([personne])
            cnx.commit()
            entity = cnx.find("Personne").one()
            self.assertEqual(entity.photo.getvalue(), b"poilu")

    def test_import_missing_required_attribute(self):
        """Check import of ext entity with missing required attribute"""
        with self.admin_access.repo_cnx() as cnx:
            importer = self.importer(cnx)
            tag = ExtEntity("Personne", 2, {"prenom": {"Jean"}})
            self.assertRaises(ValidationError, importer.import_entities, [tag])

    def test_import_inlined_relation(self):
        """Check import of ext entities with inlined relation"""
        with self.admin_access.repo_cnx() as cnx:
            importer = self.importer(cnx)
            richelieu = ExtEntity("Personne", 3, {"nom": {"Richelieu"}, "enfant": {4}})
            athos = ExtEntity("Personne", 4, {"nom": {"Athos"}})
            importer.import_entities([athos, richelieu])
            cnx.commit()
            rset = cnx.execute('Any X WHERE X is Personne, X nom "Richelieu"')
            entity = rset.get_entity(0, 0)
            self.assertEqual(entity.enfant[0].nom, "Athos")

    def test_import_non_inlined_relation(self):
        """Check import of ext entities with non inlined relation"""
        with self.admin_access.repo_cnx() as cnx:
            importer = self.importer(cnx)
            richelieu = ExtEntity("Personne", 5, {"nom": {"Richelieu"}, "connait": {6}})
            athos = ExtEntity("Personne", 6, {"nom": {"Athos"}})
            importer.import_entities([athos, richelieu])
            cnx.commit()
            rset = cnx.execute('Any X WHERE X is Personne, X nom "Richelieu"')
            entity = rset.get_entity(0, 0)
            self.assertEqual(entity.connait[0].nom, "Athos")
            rset = cnx.execute('Any X WHERE X is Personne, X nom "Athos"')
            entity = rset.get_entity(0, 0)
            self.assertEqual(entity.connait[0].nom, "Richelieu")

    def test_import_missing_inlined_relation(self):
        """Check import of ext entity with missing inlined relation"""
        with self.admin_access.repo_cnx() as cnx:
            importer = self.importer(cnx)
            richelieu = ExtEntity("Personne", 7, {"nom": {"Richelieu"}, "enfant": {8}})
            self.assertRaises(Exception, importer.import_entities, [richelieu])
            cnx.commit()
            rset = cnx.execute('Any X WHERE X is Personne, X nom "Richelieu"')
            self.assertEqual(len(rset), 0)

    def test_import_missing_non_inlined_relation(self):
        """Check import of ext entity with missing non-inlined relation"""
        with self.admin_access.repo_cnx() as cnx:
            importer = self.importer(cnx)
            richelieu = ExtEntity(
                "Personne", 9, {"nom": {"Richelieu"}, "connait": {10}}
            )
            self.assertRaises(Exception, importer.import_entities, [richelieu])
            cnx.commit()
            rset = cnx.execute('Any X WHERE X is Personne, X nom "Richelieu"')
            entity = rset.get_entity(0, 0)
            self.assertEqual(entity.nom, "Richelieu")
            self.assertEqual(len(entity.connait), 0)

    def test_import_order(self):
        """Check import of ext entity in both order"""
        with self.admin_access.repo_cnx() as cnx:
            importer = self.importer(cnx)
            richelieu = ExtEntity("Personne", 3, {"nom": {"Richelieu"}, "enfant": {4}})
            athos = ExtEntity("Personne", 4, {"nom": {"Athos"}})
            importer.import_entities([richelieu, athos])
            cnx.commit()
            rset = cnx.execute('Any X WHERE X is Personne, X nom "Richelieu"')
            entity = rset.get_entity(0, 0)
            self.assertEqual(entity.enfant[0].nom, "Athos")

    def test_update(self):
        """Check update of ext entity"""
        with self.admin_access.repo_cnx() as cnx:
            importer = self.importer(cnx)
            # First import
            richelieu = ExtEntity("Personne", 11, {"nom": {"Richelieu Diacre"}})
            importer.import_entities([richelieu])
            cnx.commit()
            rset = cnx.execute("Any X WHERE X is Personne")
            entity = rset.get_entity(0, 0)
            self.assertEqual(entity.nom, "Richelieu Diacre")
            # Second import
            richelieu = ExtEntity("Personne", 11, {"nom": {"Richelieu Cardinal"}})
            importer.import_entities([richelieu])
            cnx.commit()
            rset = cnx.execute("Any X WHERE X is Personne")
            self.assertEqual(len(rset), 1)
            entity = rset.get_entity(0, 0)
            self.assertEqual(entity.nom, "Richelieu Cardinal")


class UseExtidAsCwuriTC(TestCase):
    def test(self):
        personne = ExtEntity(
            "Personne", b"1", {"nom": {"de la lune"}, "prenom": {"Jean"}}
        )
        mapping = {}
        set_cwuri = use_extid_as_cwuri(mapping)
        list(set_cwuri((personne,)))
        self.assertIn("cwuri", personne.values)
        self.assertEqual(personne.values["cwuri"], {"1"})
        mapping[b"1"] = "whatever"
        personne.values.pop("cwuri")
        list(set_cwuri((personne,)))
        self.assertNotIn("cwuri", personne.values)
        personne = ExtEntity("Personne", "ééé", {})
        mapping = {}
        set_cwuri = use_extid_as_cwuri(mapping)
        list(set_cwuri((personne,)))
        self.assertIn("cwuri", personne.values)
        self.assertEqual(personne.values["cwuri"], {"ééé"})


class DropExtraValuesTC(CubicWebTC):
    def test(self):
        personne = ExtEntity(
            "Personne",
            b"1",
            {
                "nom": {"de la lune", "di la luna"},
                "prenom": {"Jean"},
                "enfant": set("23"),
                "connait": set("45"),
            },
        )
        log = SimpleImportLog("<unspecified>")
        list(drop_extra_values((personne,), self.schema, log))
        self.assertEqual(len(personne.values["nom"]), 1)
        self.assertEqual(len(personne.values["enfant"]), 1)
        self.assertEqual(len(personne.values["connait"]), 2)
        self.assertEqual(len(log.logs), 2)


def extentities_from_csv(fpath):
    """Yield ExtEntity read from `fpath` CSV file."""
    with open(fpath, "rb") as f:
        for uri, name, knows in ucsvreader(f, skipfirst=True, skip_empty=False):
            yield ExtEntity("Personne", uri, {"nom": {name}, "connait": {knows}})


class DataimportFunctionalTC(CubicWebTC):
    def test_csv(self):
        extenties = extentities_from_csv(self.datapath("people.csv"))
        with self.admin_access.repo_cnx() as cnx:
            store = RQLObjectStore(cnx)
            importer = ExtEntitiesImporter(self.schema, store)
            importer.import_entities(extenties)
            cnx.commit()
            rset = cnx.execute('String N WHERE X nom N, X connait Y, Y nom "Alice"')
            self.assertEqual(rset[0][0], "Bob")


if __name__ == "__main__":
    unittest_main()
